#!/bin/sh 
#
# wrap.sh: for dummy workload
# Copyleft 2002 Cliff White <cliffw@osdl.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# This is an example shell script to drive a test on the STP
# 
# The script must do these things: ( order may vary ) 
# 1. Establish the system environment. The script will be run from a test directory,
# 	created by the STP client engine. The system disk layout will be in the  
#       file ../disk.info relative to that path. A ./results directory must exist
# 	
# 2. Install and configure any application or utility code that the test needs -
#	This code should be packaged in a portable manner ( .tar.gz , .tar.bz2 ) 
# 	
# 	Examples of user code might include databases, LDAP servers
#	or any other special bits your test needs
# 3. Install and configure the actual test code. We try to keep our download sizes small,
# 	so your test code should be packaged in some fashion. 
#	Most of our standard tests are instaled and built from source.
# 4. Run the test, capuring output/results
# 5. Process the results sufficent for email and Web display - index.html must be created!
# 6. Exit cleanly. The system will be scrubbed by STP after the run. 

# Version information is a Good Thing
echo "This is a dummy test for the STP "
VERSION=0.0.1
echo "$0 version $VERSION"

# Now we know where we are 
MYHOME=`pwd`
RESDIR='./results'

# first, set up the environment. 
echo "Cleaning up the environment"
if [ -d $RESDIR ]; then
	rm -rf $RESDIR
fi
# make the results directory 
mkdir $RESDIR

# I am paranoid, i check everything - 
if [ -d $RESDIR ]; then
	echo "Directory created!"
else
	echo "Error - no results directory"
fi

# unpack the test and build it
tar xvf ./example.tar 
# This test uses a Makefile 
cd example
if [ -f ./cpu.sh ]; then
	echo "Tarball successfully unpacked"
else
	echo "Tarball failed to unpack"
	exit 1
fi
# so building it is very easy	
make
# no configuration required 
if [ -f ./cpu ]; then
	echo "Make succeeded"
else
	echo "Make failed"
	exit 1
fi

echo "Starting tests" 
# be sure to capture raw data for your results
./cpu > ../results/cpu.r
./mem > ../results/mem.r
./upt > ../results/upt.r

# get the results
# make the index.html page
# First, get to the right place
cd $MYHOME

# we have a separate script for creating our results page
# results processing could be included in this script

./results.sh


echo "All Finished!"
