#!/usr/bin/perl -w

#Programmed by Nathan Dabney of the Open Source Development Lab
#Copyright (C) 2001 Open Source Development Lab 
#
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# CHANGES
#
# 11/07/02 - Nathan Dabney <smurf@osdl.org>
# 	Added support to detect presence of STP line to avoid duplication
# 	Made vmlinuz-STP non-optional
#
# 04/22/03 - Nathan Dabney <smurf@osdl.org>
#   Handle append modifications correctly
# 07/09/03 - cliffw
#  Fix the append line, so we always have a serial cosole
#  
#   Adjust for when we already have the STP line (but don't skip)
#   Use strict

use strict;

my $append;
my $console_fix = " console=tty0 console=ttyS0,38400n8";
my $lilo = `cat /etc/lilo.conf`;

exit 0 unless ( ( -f "lilo.opt" ) || ( ! ( $lilo =~ /STP/g ) ) );

if ( -f "lilo.opt" ) {
    $append = `cat lilo.opt`;
    chomp $append;
    $append = $append . $console_fix;
    
    $lilo =~ s/^#+\s*(append=.+?)$/$1/ms;
    $lilo =~ s/append=.+?$/append="$append"/msg;
} else {
    $append = $console_fix;
}

if ( ! ( $lilo =~ /STP/g ) ) {
    $lilo .= "\nimage=/boot/vmlinuz-STP\n";
    $lilo .= "\tlabel=STP\n";
    $lilo .= "\tread-only\n";
    if ( $lilo =~ /^\s*root=(\S+)$/ms ) {
        $lilo .= "\troot=$1\n";
        $lilo .= "\tappend=\"$append\"\n";
    }
}

print "New lilo.conf:\n$lilo\n";

open (F, ">/etc/lilo.conf" ) || die "Unable to open /etc/lilo.conf: $1";
print F $lilo;
close (F);

exit 0;
