#!/usr/bin/perl -w

=head1 NAME

Eidetic STP Interface

=head1 SYNOPSIS



=head1 DESCRIPTION

Eidetic is an Internet database editor, tabularizer, indexer, and
crossreferencer. It is designed primarily to be really quick and easy to
set up and customize, with only a passing knowledge of UNIX and Perl. A
new database table and form will be able to be set up, configured, and
made available in under ten minutes. Basic configuration (introductory
text, colors, table layout, etc.) can be done through the web, with no
Perl hacking necessary. For more sophisticated customizations, a Perl
coder can override any or all of the basic functionality as needed.

Eidetic is not only a database interface, however. It provides all the
machinery needed for user input, data tabulation, administrative
editing, and reporting. Configuring a database can be as easy as
selecting the fields you want in your database, listing the fields that
should be displayed in each report, and picking the colors to display
things in.

 http://www.worldforge.org/website/tools/domesday/requirements
 http://webmake.taint.org/
 http://www.worldforge.org/website/infra/todo
 http://www.igd.fhg.de/archive/1995_www95/papers/51/WebMake/WebMake.html
 http://www.worldforge.org/website/infra/website_plan/

This is a trimmed down version of Eidetic specifically for STP

=head1 FUNCTIONS

=cut

package Eidetic::Stp;

use 5.006;
use strict;
use warnings;

require Exporter;
require DynaLoader;

use vars qw($VERSION @ISA %DEFAULTS %CONFIG);
use vars
  qw($modulename $command $query $filter $sort $maxitems $on $hilite $start $language)
  ;                                                           #CGI params
use vars qw(%tables %db_handles %modules);                    #VTables
use vars qw($db_host $db_user $db_pass);                      # Database info
use vars qw($adminAddrs $cgi $webmaster $domain $idField);    #Other stuff

our @ISA    = qw(Exporter DynaLoader);
our @EXPORT = qw( $domain
  $adminAddrs
  $webmaster
  $idField
  $eideticDir
  $srcDir
  $htmlDir
  $tmpDir
  $tmplDir
  $db_brand
  $db_host
  $db_user
  $db_pass
);
our @EXPORT_OK = qw();

use CGI;
use DBI();
use IO::File;
use POSIX qw (tmpnam strftime);
use Digest::MD5 qw(md5_hex);
use File::Copy;
use File::Spec;
use File::stat;

use constant DEBUGGING => 0;

# Eidetic.pm
$VERSION = 1.9.0;

$cgi = new CGI;
my $myself       = $cgi->self_url;
my $my_url       = $cgi->url;
my $auth_user    = $cgi->remote_user;
my $person_table = 'user';

# == Configuration ======================================================
$domain     = "osdl.org";
$adminAddrs = "bryce\@$domain";
$webmaster  = "webmaster\@$domain";
$idField    = 'uid';
my $proj_update_email = "cem\@osdl.org";
my @award_nomination_emails = ( 'bryce\@osdl.org', 'nominations\@osdl.org' );
my @project_notification_emails = (
                                    'bryce\@osdl.org',  'wookie\@osdl.org',
                                    'cliffw\@osdl.org', 'judith\@osdl.org'
);
my @project_notification_emails_jp = ( 'takazawa\@osdl.org', 'jun\@osdl.org' );

if ( $my_url eq 'http://localhost/./stp.cgi' ) {
    $my_url = "http://www.$domain/cgi-bin/stp.cgi";
}

my $limit_view_user = 1;

my $eideticDir = "/home/www/eidetic";
my $srcDir     = "/home/www/site_src";
my $incDir     = "/var/www/html/osdl_extranet/misc";
my $htmlDir    = "/home/www/site_html";
my $tmpDir     = "/home/www/site_tmp";

my $uploadDir         = "/home/www/intranet_uploads";
my $uploadMailPrefix  = "[OSDL] ";
my $uploadMailingList = "bryce\@neptune.net";           #"uploads\@$domain";

my $tmplDir  = "$eideticDir/tmpl";
my $db_brand = 'mysql';

# == End of configuration section =============

# Eventually I should probably do something more encapsulated or
# more OO (method calls?) but for now I just want to start establishing
# the parameters listed out
$modulename = $cgi->param( 'modulename' ) || 'users';
$command    = $cgi->param( 'command' )    || 'userpage';
$query      = $cgi->param( 'query' )      || '';
$filter     = $cgi->param( 'filter' )     || '';
$sort       = $cgi->param( 'sort' )       || '';
$maxitems   = $cgi->param( 'maxitems' )   || 25;
$start      = $cgi->param( 'start' )      || 0;
$on         = $cgi->param( 'on' )         || '';
$hilite     = $cgi->param( 'hilite' )     || '';
$language = lc( $cgi->param( 'language' ) ) || 'en';

if ( $cgi->server_software ne 'cmdline' && $command ne 'redirect' ) {
    print $cgi->header();
}

# Permission levels
my $WORLD      = 0;
my $VALID_USER = 1;
my $RESOURCE   = 5;
my $OWNER      = 10;
my $FIELD      = 50;
my $ADMIN      = 100;

# VTables
#my %tables;                     # All table info
#my %db_handles;                 # Cache of database handles
#my %modules;
my %storageError;
my @colors;
my %colorScheme;
my $curColor = 0;

my %eideticDefaults = (
                        modulename => '',
                        command    => '',
                        on         => '',
                        start      => 0,
                        query      => '',
                        filter     => '',
                        sort       => '',
                        maxitems   => 25,
                        hilite     => '',
                        page       => 0
);

########################################################################
## MODULARITY SUBSYSTEM                                               ##
########################################################################
# This will be Eidetic::Modularity

=head1 Modularity subsystem

=head2 registerModule(NAME, PARAMS)

Registers a module into Eidetic's virtual table.

=cut

sub registerModule($$) {

    # This command adds a function to the vtable
    my ( $modulename, $params ) = @_;

    $modules{ $modulename } = $params;

    # Is there a way to force/require what parameters are included
    # when the module is registered?
}

=head2 getActions(MODULE, COMMAND)

Gets a list of actions availble for the particular command in the
module.

=cut

sub getActions($$) {
    my ( $modulename, $command );

    # TODO:  Need to check if the various hashes are defined

    return '';    #$modules{$modulename}->{$command}->{actions};
}

=head2 dispatch(CGI, MODULENAME, COMMAND, USER, DATA)

Sends a command out for execution

=cut

sub dispatch($$$$$$) {
    my ( $cgi, $modulename, $command, $on, $user, $input_data ) = @_;
    my ( $content, %form );
    my $dbh      = $db_handles{ 'eidetic' };
    my @emails   = ();
    my $commands = $modules{ $modulename }->{ commands };
    my @tables   = $modules{ $modulename }->{ tables } || ( '' );
    my ( $database, $table ) = split /\:/, $tables[ 0 ];
    die "Invalid modulename; can't find matching database" unless $database;
    die "Invalid modulename; can't find matching table"    unless $table;

    my %defaults;
    if ( $modules{ $modulename }->{ defaults } ) {
        %defaults = $modules{ $modulename }->{ defaults };
    }
    $user->{ user_id } = $user->{ $idField };

    if ( $modulename eq 'admin' || $modulename eq 'stp' ) {
        my $handler = $modules{ $modulename }->{ handler };

        &$handler( $cgi, $modulename, $command, $on, $user, $input_data );
        return;
    } elsif ( exists $commands->{ $command } ) {

        # Can I abstract this into a multi-command processor?
        die "Unknown command $command"
          unless ( exists $commands->{ $command } );
        my ( $newCommand, $newQuery, $noticeMsg, $newOn ) =
          $commands->{ $command }->(
          $modulename, $command,    $table,   $dbh,
          $on,         $input_data, \@emails, $command,
          $query,      $user,       ''
        );
        die "Unknown command $newCommand"
          unless ( exists $commands->{ $newCommand } );
        if ( $newOn && $newOn ne '' ) { $on = $newOn; }
        if ( $newCommand ne $command || $newQuery ne $query ) {
            if ( $command eq 'remove' ) {
                $on = '';
            }    # Prevent trying to display the just-deleted item
            $commands->{ $newCommand }->(
              $modulename, $newCommand, $table,   $dbh,
              $on,         $input_data, \@emails, '',
              '',          $user,       $noticeMsg
            );
        }
        return;

    } else {

        # Add error message about unfound module
    }
}

=head2 getfrom_db(DBH, TABLE, START, FILTER, SORT, FIRST, QTY)

Gets specified rows from DBH, using selectRecords, and returns a handle
for the first row found.

=cut

my $getfrom_db_count = 0;

sub getfrom_db {
    my ( $dbh, $table, $on, $filter, $sort, $first, $qty ) = @_;

    my $rightify = ( $on =~ /\!/ );
    my $newFilter = parseOn( $on );

    if ( $filter ) {
        $filter .= " AND ";
    }
    $filter .= "$idField $newFilter";

    my $handle = selectRecords( $dbh, $table, $filter, $sort, $first, $qty );

    if ( $handle == 0 ) {

        # No available records were found
        die
"No available records were found, so we couldn't process the request.";
    }

    $getfrom_db_count = 0;
    if ( $rightify && $handle->rows > $maxitems ) {
        $getfrom_db_count = $maxitems - $handle->rows;
    }
    while ( $getfrom_db_count < 0 ) {

        # throw away unwanted initial elements
        my $data = $handle->fetchrow_hashref;
        $getfrom_db_count++;
    }

    return $handle;
}

=head2 getnext_db(HANDLE)

Takes $handle returned from getfrom_db and returns the next row or C<undef>

=cut

# Returns a reference to a hash containing the data item
sub getnext_db($) {
    my ( $handle ) = @_;

    $getfrom_db_count++;
    if ( $getfrom_db_count > $maxitems ) {
        return undef;    # We have reached the maximum number of items we want
    }

    my $data = $handle->fetchrow_hashref;
    return $data;
}

=head2 putin_tmpl(STYLE, DATA)

Returns the template read from $style, with substitution using %data

=cut

sub putin_tmpl {
    my ( $style, %data ) = @_;
    my $fname = $style;
    return substituteValues( loadFile( $fname ), \%data, '&nbsp;' );
}

=head2 sendto_screen(TO, TITLE, CONTENT, LEFT_COLUMN, RIGHT_COLUMN)

Substitutes TITLE-R_C, $srcDir/header.inc as header, and
$srcDir/footer.tmpl as footer into $srcDir/main.tmpl, then C<print>s it.

=cut

sub sendto_screen($$$;$$) {
    my ( $to, $title, $content, $left_column, $right_column ) = @_;

    # $left_column is a placeholder for now, until the link box editor
    # is in place.  Perhaps the column text could be autogenerated in
    # this routine?

    my $tmplText = loadFile( "$srcDir/main.tmpl" );
    my $header   = loadFile( "$incDir/stp_header.html" );
    my $footer   = loadFile( "$incDir/stp_footer.html" );

    if ( !$left_column )  { $left_column  = ''; }
    if ( !$right_column ) { $right_column = ''; }

    $tmplText =~ s/\[\%\s*header\s*\%\]/$header/g;
    $tmplText =~ s/\[\%\s*footer\s*\%\]/$footer/g;
    $tmplText =~ s/\[\%\s*title\s*\%\]/$title/g;
    $tmplText =~ s/\[\%\s*modulename\s*\%\]/$modulename/g;
    $tmplText =~ s/\[\%\s*left_column\s*\%\]/$left_column/g;
    $tmplText =~ s/\[\%\s*content\s*\%\]/$content/g;
    $tmplText =~ s/\[\%\s*right_column\s*\%\]/$right_column/g;

    print $tmplText;
}

=head2 getfrom_file(FILE)

Returns hashref with $file loadFile ed into {content}.

=cut

sub getfrom_file {
    my %data;
    $data{ content } = loadFile( @_ );
    return \%data;
}

=head2 getfrom_cgi(FROM)

Returns anon hashref.

=cut

sub getfrom_cgi($) {
    my $from = @_;
    my %data;
    return \%data;
}

=head2 parseOn(ON)

Parses ON into SQL. Syntax: $firstId-$lastId means rows with ids between $firstId
and $lastId; if either is omitted, default is 0 or largest id, respectively. Will
also do joins eventually.

=cut

sub parseOn($) {
    my @a = split /\,/, $_[ 0 ];
    my ( @ins, @betweens, $firstId, $lastId );
    foreach ( @a ) {
        if ( /^\-(\d+)\!/ ) {
            push @betweens, " BETWEEN 0 AND $1";
        } elsif ( /^(\d+)-$/ ) {
            $firstId = $1;
            push @betweens, " >= $firstId";
        } elsif ( /\-/ ) {
            ( $firstId, $lastId ) = split /\-/;
            push @betweens, " BETWEEN $firstId AND $lastId";
        } elsif ( !/\-/ ) {
            push @ins, $_;
        } else {

            # syntax error; ignore.
        }
    }

    my ( $ins );
    if ( $#ins >= 0 ) {
        $ins = "IN (" . join ( ', ', @ins ) . ") ";
        return join ( " OR $idField ", ( $ins, @betweens ) );
    } else {
        return join ( " OR $idField ", @betweens );
    }
}

########################################################################
## DATABASE SUBSYSTEM                                                 ##
########################################################################

=head1 Database subsystem

=head2 prepareDatabase(MODULENAME)

Given a valid module name, this initializes all the databases that it
will need.

=cut

sub prepareDatabases($) {
    my ( $modulename ) = @_;

    # Prepare the databases
    my $dbname;

    # Connect to databases as required
    foreach $dbname ( $modules{ $modulename }->{ 'databases' } ) {

        # This is a little kludgy and needs cleaning up

        if ( $dbname ne '' ) {
            $db_handles{ $dbname } = DBI->connect(
                                                   $db_host, $db_user,
                                                   $db_pass
            );
            if ( !$db_handles{ $dbname } ) {
                die
                  "failure Could not connect to database $db_host as $db_user";

                # TODO:  Add better error handling
            }
        }
    }
}

=head2 getDbHandle(DBNAME)

Returns a handle to db DBNAME; throws on failure.

=cut

sub getDbHandle($) {
    my ( $dbname ) = @_;
    if ( exists $db_handles{ $dbname } ) {
        return $db_handles{ $dbname };
    } else {
        $db_handles{ $dbname } = DBI->connect( $db_host, $db_user, $db_pass );
        if ( !$db_handles{ $dbname } ) {
            die "failure Could not connect to database";
        }
        return $db_handles{ $dbname };
    }
}

=head2 createDatabaseTable(DBH, TABLE, SCHEMA)

Creates $table in $dbh, with fields and types given by: $schema{$field}=$type.
Probably {mysql,pg}-specific?

=cut

sub createDatabaseTable($$%) {
    my ( $dbh, $table, %schema ) = @_;

    if ( !$dbh ) {
        die "Cannot connect to database : $!";
    }

    if ( $table eq '' ) {
        die "Invalid table name.";
    }

    my $cmd = "CREATE TABLE $table (";

    my $sep = '';
    while ( my ( $field, $type ) = each %schema ) {
        $cmd .= "$sep $field $type";
        $sep = ',';
    }
    $cmd .= ')';

    # Execute the SQL statement  - should I use do?
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command: " . $dbh->errstr;
    my $rows_affected = $sth->execute();

    # Perhaps also create an INDEX?
}

=head2 addDatabaseFields(DBH, TABLE, SCHEMA)

Like createTable but doesn't create anything.' DB-specific?

=cut

sub addDatabaseFields($$%) {
    my ( $dbh, $table, %schema );

    # TODO: verify the field doesn't already exist, and is a valid field name

    my $cmd = "ALTER TABLE $table ADD (";

    my $sep = '';
    while ( my ( $field, $type ) = each %schema ) {
        $cmd .= "$sep $field $type";
        $sep = ',';
    }
    $cmd .= ')';

    # Execute the SQL statement  - should I use do?
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command: " . $dbh->errstr;
    my $rows_affected = $sth->execute();
}

=head2 closeDatabases( )

This routine simply closes whatever databases are currently opened.

=cut

sub closeDatabases() {
    my $dbname;

    # We're done with the db, so disconnect it
    foreach $dbname ( keys %db_handles ) {
        $db_handles{ $dbname }->disconnect();
    }
}

=head2 authorizeUser(DBH, ON, USERNAME)

Return true if user #$on = HTTP-authorized user, false if not.

=cut

sub authorizeUser($$$) {
    my ( $dbh, $on, $username ) = @_;

    if ( $on !~ /^\d+$/ ) {
        return 0;
    }
    if ( !isInDatabase( $dbh, $person_table, "$idField='$on'" ) ) {
        return 0;
    }

    my $sth = selectRecords( $dbh, $person_table, "$idField=$on", '', 0, 1 );
    if ( $sth == 0 ) {
        return 0;
    }
    my $user = $sth->fetchrow_hashref;

    if ( $user->{ descriptor } ne $auth_user ) {
        return 0;
    }

    return 1;
}

=head2 checkMissingElements(DBH, TABLE, DATA)

# Given a table name and reference to a data hash
# this routine fills in defaults and returns a list of
# missing elements.

=cut

sub checkMissingElements($$$) {
    my ( $dbh, $table, $data ) = @_;
    my @missingElements;
    my $schemaDefault = getTableInfo( $dbh, $table, 'DEFAULTS' );

    if ( !$dbh ) {
        die "Internal error in checkMissing Elements:  $dbh not defined";
    }

    if ( !$table ) {
        die "Internal error in checkMissing Elements:  $table not defined";
    }

    if ( !$data ) {
        die "Internal error in checkMissing Elements:  $data not defined";
    }

    foreach my $key ( keys %{ $schemaDefault } ) {
        if ( !exists $data->{ $key } ) {
            if ( !$schemaDefault->{ $key } ) {
                $data->{ $key } = '';
            } elsif ( $schemaDefault->{ $key } eq 'required' ) {
                push ( @missingElements, $key );
            } else {
                $data->{ $key } = $schemaDefault->{ $key };
            }
        }
    }
    return @missingElements;
}

=head2 checkInvalidElements($$$)

Returns empty list

=cut

sub checkInvalidElements($$$) {
    my @invalidElements;
    return @invalidElements;
}

=head2 getFieldText(DBH, DATA, SOURCE, FIELD)

Gets the value for a given field from a given source.  This is what
parses the Eidetic-style fields to give autoinclusion of database
parameters.

=cut

sub getFieldText {
    my ( $dbh, $data, $source, $field ) = @_;

    my $table = $source;
    if ( $source =~ /^\w+\_$idField\_(\w+)$/ ) {
        $table = $1;
    } elsif ( $source =~ /^\w+\_by$/ || $source =~ /^\w+\_to$/ ) {
        $table = $person_table;
    }
    my $sourceId = 0;
    if ( exists $data->{ "$source\_$idField" } ) {
        $sourceId = $data->{ "$source\_$idField" };
    } elsif ( exists $data->{ $source } ) {
        $sourceId = $data->{ $source };
    } elsif ( exists $data->{ lc "$source\_$idField" } ) {
        $sourceId = $data->{ lc "$source\_$idField" };
    } elsif ( exists $data->{ lc $source } ) {
        $sourceId = $data->{ lc $source };
    }
    my $sth = selectRecords( $dbh, lc( $table ), "$idField=$sourceId", '' );
    my $item = $sth->fetchrow_hashref;
    return $item->{ $field };
}

=head2 insertRecord(DBH, TABLE, DATA)

insertRecord - Inserts data into a db table, assuming it meets schema

=cut

sub insertRecord($$%) {
    my ( $dbh, $table, %data ) = @_;
    my ( $i );

    # For this table, determine what the columns are
    my @fields = @{ getTableInfo( $dbh, $table, 'STORE_FIELDS' ) };

    # Assemble the SQL statement
    my $cmd = "INSERT INTO $table (" . join ( ',', @fields ) . ") VALUES (";

    my %type = %{ getTableInfo( $dbh, $table, 'TYPES' ) };
    for ( $i = 0; $i < ( $#fields ); $i++ ) {
        $cmd .=
          formatItem( $data{ $fields[ $i ] }, $type{ $fields[ $i ] } ) . ",";
    }
    $cmd .=
      formatItem( $data{ $fields[ $#fields ] }, $type{ $fields[ $#fields ] } )
      . ")";

    # Execute the SQL statement 
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command: " . $dbh->errstr;
    my $rows_affected = $sth->execute();

    if ( $rows_affected == 0 ) {
        die "Insert failed for $cmd $!: " . $dbh->errstr;
    }

    # retrieve any autoincrement that the database created

    if ( $sth->{ mysql_insertid } ) {
        return $sth->{ mysql_insertid };
    } else {
        return $sth->{ insertid };
    }
}

=head2 changePermission(DBH, UID, RESOURCE_TAG, ACCESS)

Changes access to a resource for a given user. It returns 1 on success, throws
on failure.

=cut

sub changePermission($$$$) {
    my ( $dbh, $user_uid, $resource_tag, $access ) = @_;

    my $cmd = 'INSERT INTO permission ';
    $cmd .= '(user_uid, resource_tag, access, trust_score, rsf)';
    $cmd .= "VALUES ($user_uid, '$resource_tag', '$access', 100, 1)";

    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command: " . $dbh->errstr;

    return $sth->execute();
}

=head2 isInDatabase(DBH, TABLE, FILTER)

This routine searches to see if any records match the given filter in the
given table within the database handle provided.  It returns
true if there are one or more records matching the criteria.

=cut

sub isInDatabase($$$) {
    my ( $dbh, $table, $filter ) = @_;

    die "No table given for isInDatabase(,,$filter)" unless $table;
    my @fields = @{ getTableInfo( $dbh, $table, 'FIELDS' ) };

    my $cmd = "SELECT $fields[0] FROM $table";

    # Need WHERE condition based on $filter, and need to do sorting
    if ( $filter ne '' ) {
        $cmd .= " WHERE $filter";    # For now we'll use plain old SQL
    }

    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command $cmd " . $dbh->errstr();
    my $numRows = $sth->execute;
    if ( $sth->errstr() ) {
        die "failure Cannot execute database command $cmd" . $sth->errstr();
    } elsif ( !$numRows ) {
        die "failure Cannot execute database command $cmd";
    }

    return $numRows > 0;
}

=head2 selectRecords(DBH, TABLE, FILTER, SORT)

Begins a query of the database using the specified filter and sort,
and returns the query's statement handle.  Use $queryHandle->fetchall_arrayref,
$queryHandle->fetchrow_hashref, or etc. as desired to get the data.

=cut

sub selectRecords {
    my ( $dbh, $table, $filter, $sort, $first, $qty ) = @_;

    die
"selectRecords called with no table specified, for filter='$filter', sort='$sort'"
      unless $table;
    my @fields;
    my $junction = '';
    my $left     = 0;

    if ( !$first ) { $first = 0; }
    if ( !$qty )   { $qty   = -1; }

    # This figures out junction table joining
    ( $table, $junction ) = split /\<\-/, $table;

    if ( $junction && $junction ne '' ) {
        if ( $junction =~ /^\+(.*)$/ ) {
            $junction = $1;
            $left     = 1;
        }
        my @junction_fields = @{ getTableInfo( $dbh, $junction, 'FIELDS' ) };
        my @target_fields   = @{ getTableInfo( $dbh, $table,    'FIELDS' ) };
        foreach my $target_field ( @target_fields ) {
            push @fields, "$table.$target_field";
        }
        foreach my $junction_field ( @junction_fields ) {
            push @fields, "$junction.$junction_field";
        }
    } else {
        @fields = @{ getTableInfo( $dbh, $table, 'FIELDS' ) };
    }
    my $cmd = "SELECT " . join ( ', ', @fields ) . " FROM $table";

    # Need WHERE condition based on $filter, and need to do sorting
    if ( $left ) {

        #        $cmd .= " LEFT JOIN $junction ON $junction.";
        #        $cmd .= ucfirst ($table);
        #        $cmd .= "_$idField = $table.$idField";
        $cmd .=
" LEFT JOIN $junction ON $junction.$table\_$idField = $table.$idField";

    } elsif ( $junction ) {
        $cmd .= ", $junction";
        $cmd .= " WHERE $junction.rsf<>0 AND $table.rsf<>0 AND $junction.";

        #        $cmd .= ucfirst ($table);
        #        $cmd .= "_$idField = $table.$idField";
        $cmd .= $table . "_$idField = $table.$idField";
    } else {
        $cmd .= " WHERE rsf<>0";
    }
    if ( $filter ) {
        $cmd .= " AND ($filter)";
    }
    if ( $sort =~ /\!/ ) {
        my $s = $sort;
        $s =~ s/\!//g;
        $cmd .= " ORDER BY $s DESC";    # For now we'll use plain old SQL
    } elsif ( $sort ) {
        $cmd .= " ORDER BY $sort";
    } elsif ( $junction ) {

        # By default, we sort by the sub-table's uid
        $cmd .= " ORDER BY $table\_$idField";
    }
    if ( $first > 0 && $qty > 0 ) {
        $cmd .= " LIMIT $first,$qty";
    }
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command $cmd" . $dbh->errstr();
    my $numRows = $sth->execute
      or die "failure Cannot execute database command $cmd" . $sth->errstr();

    return $sth;
}

=head2 getDatabases(DBH, SERVER)

Returns a list of available databases in the db server

=cut

# getDatabases
# Returns a list of available databases in the db server
sub getDatabases($$) {
    my ( $dbh, $server ) = @_;
    my @databases = $dbh->func( $server, '_ListDBs' );

    if ( not @databases ) {

        # emit error
    }
    return @databases;
}

=head2 getTables(DBH, SERVER)

Returns a list of available tables in a database

=cut

# getTables
# Returns a list of available tables in a database
sub getTables($$) {
    my ( $dbh, $server ) = @_;
    my @tables = $dbh->func( $server, '_ListTables' );

    if ( not @tables ) {

        # emit error
    }
    return @tables;
}

=head2 checkTableExistance(DBH, TABLE)

Returns true if the table exists, false if not.

=cut

sub checkTableExistance($$) {
    my ( $dbh, $table ) = @_;
    my $cmd = "SHOW FIELDS FROM $table";
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database statement: " . $dbh->errstr;
    if ( $sth->execute() ) {
        return 1;
    } else {
        return 0;
    }
}

=head2 getTableInfo(DBH, TABLE, INFOTYPE)

This routine retrieves a data structure which describes a particular table
and is a real workhorse for Eidetic.  It returns a reference to an array or
hash for the $infotype.

=cut

# getTableInfo
# This routine retrieves a data structure which describes a particular table
# and is a real workhorse for Eidetic.  It returns a reference to an array or
# hash for the $infotype.  It requires and assumes that the table actually 
# exists in the database, and if it doesn't, it dies.
# Thus call checkTableExistance($dbh, $table) to verify first
sub getTableInfo($$$) {
    my ( $dbh, $table, $infotype ) = @_;

    die "Table name not specified for '$infotype'" unless ( $table ne '' );

    if ( !$dbh ) {
        die "Database not active for reading $table";
    }

    if ( !exists $tables{ $table } ) {
        my $cmd = '';

        if ( $db_brand eq 'mysql' ) {
            $cmd = "SHOW FIELDS FROM $table";
        } elsif ( $db_brand eq 'postgresql' ) {

            # PostGRES quick-o-hack!
            $cmd = qq| SELECT
   a.attname,
   a.attrelid,
   format_type(a.atttypid, a.atttypmod),
   a.attnotnull,
   a.atthasdef,
   (
     SELECT 'PRI' from pg_index i, pg_attribute a2
     WHERE c.relname = '$table'
       AND i.indrelid = c.oid
       AND i.indexrelid = a2.attrelid
       AND a.attrelid = c.oid
       AND a2.attname = a.attname
   ),
   (
     SELECT substring(d.adsrc for 128)
     FROM pg_attrdef d, pg_class c
     WHERE c.relname = '$table'
       AND c.oid = d.adrelid
       AND d.adnum = 1
       AND a.atthasdef = 't'
   )
   FROM pg_attrdef d, pg_class c, pg_attribute a
   WHERE c.relname = '$table'
     AND a.attnum > 0
     AND a.attrelid = c.oid
     AND d.adrelid = c.oid
 |;

        } else {
            die "Unsupported database brand";
        }

        my $sth = $dbh->prepare( $cmd )
          or die "failure Cannot prepare database statement: " . $dbh->errstr;
        my $out = $sth->execute()
          or die "failure Cannot execute database statement: " . $sth->errstr;

        my %permissions;
        if ( $modulename eq 'users' ) {
            $permissions{ display } = $OWNER;
            $permissions{ submit }  = $OWNER;
            $permissions{ change }  = $OWNER;
            $permissions{ remove }  = $ADMIN;
        } else {
            $permissions{ display } = $WORLD;
            $permissions{ submit }  = $WORLD;
            $permissions{ change }  = $WORLD;
            $permissions{ remove }  = $WORLD;
        }
        my ( %descs, %types, %nulls, %keys, %defaults, %extra, %access );
        my ( @fields, @input_fields, @output_fields, @store_fields );
        while ( my @table_info = $sth->fetchrow_array ) {
            my ( $field, $type, $null, $key, $default, $extra );

            if ( $db_brand eq 'mysql' ) {
                ( $field, $type, $null, $key, $default, $extra ) = @table_info;
            } elsif ( $db_brand eq 'postgresql' ) {
                my $foo;
                ( $field, $foo, $type, $null, $key, $default ) = @table_info;
            } else {
                die "Unsupported brand of database";
            }

            push @fields, $field;

            $descs{ $field } = ucfirst( $field );
            $descs{ $field } =~
              s/\_(\w)/ \U$1/g;    # Convert _ to ' ' & title-case
            $types{ $field }    = $type;
            $nulls{ $field }    = $null;
            $keys{ $field }     = $key;
            $defaults{ $field } = $default;
            $extra{ $field }    = $extra;

            # To be added later if/when needed
            #           if (1==2) {
            #               # If we are doing field level access control...
            #               $access{$field}   = $WORLD;
            #           }

            if ( $field ne $idField ) {
                push @store_fields, $field;
            }
            if ( $field ne $idField
                 && $field ne 'rsf'
                 && $field !~ /^created/
                 && $field !~ /^last\_updated/ )
            {
                push @input_fields, $field;
            }
            push @output_fields, $field;
        }

        $tables{ $table } = {
                              PERMISSIONS   => \%permissions,
                              RESOURCE_NAME => '',
                              FIELDS        => \@fields,
                              STORE_FIELDS  => \@store_fields,
                              INPUT_FIELDS  => \@input_fields,
                              OUTPUT_FIELDS => \@output_fields,
                              FIELD_ACCESS  => \%access,
                              DESCS         => \%descs,
                              TYPES         => \%types,
                              NULLS         => \%nulls,
                              KEYS          => \%keys,
                              DEFAULTS      => \%defaults,
                              EXTRA         => \%extra
        };

        # RESOURCE_NAME is used to group tables that are part of a
        # system.
    }

    if ( exists $tables{ $table }->{ $infotype } ) {
        return $tables{ $table }->{ $infotype };
    } else {
        die "failure Could not get '$infotype' for '$table'";
    }
}

=head2 checkType(TYPE, ITEM)

This routine verifies that the given item is of the type specified

=cut

# checkType
# This routine verifies that the given item is of the type specified
sub checkType($$) {
    my ( $type, $item ) = @_;

    # Later, should I change this into a hash of types to fptrs,
    # where each function validates a particular data type, so
    # users of this function can add custom type checking?
    if ( !( $type )
        || ( $type =~ /^tinyint/i && $item =~ /^[+-]?\d+$/ )
        || ( $type =~ /^int/i     && $item =~ /^[+-]?\d+$/ )
        || ( $type =~ /^float/i
             && $item =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/ )
        || ( $type =~ /^varchar\((\d+)\)$/i && length $item < $1 )
        || ( $type =~ /^date/i && $item =~ /^\d\d\d\d-\d\d-\d\d$/ )
        || ( $type eq 'text' )
        || ( $type eq 'string' ) )

      # Need to add soon:  tinyint, date, text
    {
        return 1;
    } else {
        return "Bad type $type specified.";
    }
}

=head2 formatItem(ITEM, FORMAT)

Puts data item into the format specified

=cut

# Puts data item into the format specified
sub formatItem($$) {
    my ( $item, $format ) = @_;

    $item = trim( $item );
    $item =~ s/\"/\\\"/g;
    $item =~ s/\'/\\\'/g;

    if ( !defined $item ) {
        $item = '';
    }
    if ( !defined $format ) {
        $format = '';
    }

    # Later I should add a more sophisticated formatting system
    # Perhaps do a lookup of the format from the database?
    # I'd need to load the database with default formats, of course...
    if ( $format eq 'string' || $format =~ /char/i || $format =~ /text/i ) {
        return "'" . $item . "'";
    } elsif ( $format =~ /enum/ ) {
        return "'" . $item . "'";
    } elsif ( $format eq 'date' || $format eq 'datetime' ) {
        return "'" . $item . "'";
    } elsif ( $item eq '' ) {
        return '0';
    } else {
        return $item;
    }
}

=head2 validateAndStoreRecord(MODULE, USERID, DATA)

#Validates and stores %data. Doesn't check yet for duplicate or invalide 
elements.

=cut

sub validateAndStoreRecord($$%) {
    my ( $modulename, $user_id, %data ) = @_;

    my ( $dbt, $table, $dbname, @missingElements, @invalidElements, $rec_id );
    my @dbts = $modules{ $modulename }->{ 'tables' };

    if ( $modulename eq 'tests' || $modulename eq 'host_types' ) {
        push @dbts, 'eidetic:test_to_host_type';
    }
    if ( $modulename eq 'projects' || $modulename eq 'hosts' ) {
        push @dbts, 'eidetic:project_to_host_junction';
        push @dbts, 'eidetic:project_to_host';
    }
    if ( $modulename eq 'projects' || $modulename eq 'users' ) {
        push @dbts, 'eidetic:project_to_user_junction';
        push @dbts, 'eidetic:project_to_person';
    }
    foreach $dbt ( @dbts ) {
        ( $dbname, $table ) = split /:/, $dbt;

        my $dbh = getDbHandle( $dbname );

# If this table has a uniqueness field, make sure we're not adding a duplicate
        next unless checkTableExistance( $dbh, $table );
        my %keys = %{ getTableInfo( $dbh, $table, 'KEYS' ) };

        my $filter = '';
        my $keyfield;
        foreach $keyfield ( keys %keys ) {
            next
              if ( !$keyfield
                   || !exists $keys{ $keyfield }
                   || !exists $data{ $keyfield } );
            if ( $keys{ $keyfield } eq 'MUL' || $keys{ $keyfield } eq 'PRI' ) {
                $filter =
                  "$keyfield='$data{$keyfield}'"
                  ;    # assuming string here - could be anything
                if ( isInDatabase( $dbh, $table, $filter ) ) {

# Indicate to the user that this record already appears to be in there
                }
            }
        }

        my $sth = $dbh->prepare( "SELECT NOW()" );
        $sth->execute;
        my $now = $sth->fetchrow;

        # Add the housekeeping data elements
        $data{ created_by }        = $user_id;
        $data{ created_date }      = $now;
        $data{ last_updated_by }   = $user_id;
        $data{ last_updated_date } = $now;
        $data{ rsf }               = 1;

        # Verify we have the required elements; if not, make a list.
        # This also fills in defaults where appropriate.
        push ( @missingElements, checkMissingElements( $dbh, $table, \%data ) );

# Next verify the data is in the right formats, and do conversions as appropriate
        push ( @invalidElements, checkInvalidElements( $dbh, $table, \%data ) );

        # Update junction tables...
        if ( $table =~ /_to_/ ) {
            updateJunctions( $dbh, $modulename, $table, %data );
        } else {
            $data{ $idField } = insertRecord( $dbh, $table, %data );
        }
    }

    return %data;
}

=head2 updateJunctions(DBH, MODULE, TABLE, DATA)

Update junction tables.

=cut

sub updateJunctions($$$%) {
    my ( $dbh, $modulename, $table, %data ) = @_;
    my $id = $data{ $idField };

    # This indicates a junction table
    my ( $table1, $table2 ) = split /\_to\_/, $table;
    $table2 =~ s/\_junction$//;

    # The next two if statements are to help us transition from user to person
    if ( $table1 eq 'person' or $table1 eq 'user' ) {
        $table1 = $person_table;
    }

    if ( $table2 eq 'person' or $table2 eq 'user' ) {
        $table2 = $person_table;
    }

    my ( $source, $target ) = ( '', '' );
    if ( $modulename eq $table1 . "s" ) {    # ugly kludge...
        $source = "$table1\_$idField";
        $target = "$table2\_$idField";
    } elsif ( $modulename eq $table2 . "s" ) {
        $target = "$table1\_$idField";
        $source = "$table2\_$idField";
    } else {
        die
"Could not figure out source and target from $modulename, $table1, and $table2";
    }

    my %elements;

    # Extract all the elements that need adding to the db
    while ( my ( $field, $value ) = each %data ) {
        if ( $field =~ /^$table\:\:$target\[/ ) {
            $elements{ $value } = 1;    # Mark it
        }
    }

    my $cmd = "SELECT $source, $target, rsf FROM $table WHERE $source = $id";
    my $sth = $dbh->prepare( $cmd )
      or die "failure Cannot prepare database command: " . $dbh->errstr;
    $sth->execute;
    my ( @inserts, @deletes, @undeletes );
    while ( my $data = $sth->fetchrow_hashref ) {
        if ( !exists $elements{ $data->{ $target } } ) {

            # Mark this one for deletion
            push @deletes, $data->{ $target };
            delete $elements{ $data->{ $target } };
        } elsif ( $elements{ $data->{ $target } } == 1 ) {

            # undelete if necessary
            if ( $data->{ rsf } < 1 ) {
                push @undeletes, $data->{ $target };

                # Else we need do nothing to it
            }
            delete $elements{ $data->{ $target } };
        }
    }

    # If any elements are left, INSERT them
    @inserts = keys %elements;

    my $item;

    # Next perform the database commands to do all of these updates
    foreach $item ( @inserts ) {
        $cmd = "INSERT INTO $table ($source, $target, rsf) ";
        $cmd .= "VALUES ($id, $item, 1)";

        $sth = $dbh->prepare( $cmd )
          or die "failure Cannot prepare database command: " . $dbh->errstr;
        $sth->execute;
    }

    foreach $item ( @deletes ) {
        $cmd = "DELETE FROM $table WHERE ";
        $cmd .= "$source=$id AND $target=$item ";

        $sth = $dbh->prepare( $cmd )
          or die "failure Cannot prepare database command: " . $dbh->errstr;
        $sth->execute;
    }

    foreach $item ( @undeletes ) {
        $cmd = "UPDATE $table SET rsf=1 WHERE ";
        $cmd .= "$source=$id AND $target=$item ";

        $sth = $dbh->prepare( $cmd )
          or die "failure Cannot prepare database command: " . $dbh->errstr;
        $sth->execute;
    }
}

=head2 updateRecord(MODULE, USER, DATA)

#Updates $module's tables with %data.

=cut

sub updateRecord($$%) {
    my ( $modulename, $user_id, %data ) = @_;

    my $id = $data{ $idField };
    if ( $id !~ /^\d+$/ ) {
        die "$idField '$id' not specified correctly in updateRecord";
    }

    my ( $dbt, $table, $dbname, @missingElements, $rec_id );
    my @dbts = ( $modules{ $modulename }->{ 'tables' } );
    if ( $modulename eq 'tests' || $modulename eq 'host_types' ) {
        push @dbts, "eidetic:test_to_host_type";
    }
    foreach $dbt ( @dbts ) {
        ( $dbname, $table ) = split /\:/, $dbt;
        my $dbh = getDbHandle( $dbname );

        my %type = %{ getTableInfo( $dbh, $table, 'TYPES' ) };
        my $sep  = '';

        if ( $table =~ /_to_/ ) {
            updateJunctions( $dbh, $modulename, $table, %data );

        } else {
            my $sth = $dbh->prepare( "SELECT NOW()" );
            $sth->execute;
            my $now = $sth->fetchrow;

            # Add the housekeeping data elements
            $data{ last_updated_by }   = $user_id;
            $data{ last_updated_date } = $now;

            my $cmd = '';

            foreach my $field ( keys %data ) {
                if ( $field eq $idField ) {

                    # skip it.
                } elsif ( exists $type{ $field } && exists $data{ $field } ) {
                    if ( $cmd eq '' ) { $cmd = "UPDATE $table SET "; }
                    $cmd .= $sep . $field . '='
                      . formatItem( $data{ $field }, $type{ $field } );
                    $sep = ', ';
                } else {

# User requested a change to a field that we know nothing about
                }
            }

            if ( $cmd ne '' ) {
                $cmd .= " WHERE $idField=$id";

                # Execute the SQL statement  - should I use do?
                $sth = $dbh->prepare( $cmd )
                  or die "failure Cannot prepare database command: "
                  . $dbh->errstr;
                my $rows_affected = $sth->execute();

                if ( $rows_affected == 0 ) {
                    die "Update failed for $cmd $!: " . $sth->errstr;
                }
            }
        }
    }
    return %data;
}

########################################################################
## COMMAND SYSTEM                                                     ##
########################################################################

=head1 Command subsystem

=head2 Standard prototype

A standard command sub takes $modulename, $command, $table, $dbh, $on,
$dataHandle (hashref to CGI params), $emails, $nextCommand, $nextQuery,
$defaults, $noticeMsg, and returns list of $nextCommand, $nextQuery,
$noticeMsg, and (optionally) $on.

#=head2 do_intro
#
#Displays $srcDir/$modulename/display.intro.page.tmpl or a simple intro message.
#
#=cut
#sub do_intro {
#    my ($modulename, $command, $table, $dbh, $on, $dataHandle, $emails,
#        $nextCommand, $nextQuery, $defaults, $noticeMsg) = @_;
#
#    my $content;
#    my $filename = "$srcDir/$modulename/display.intro.page.tmpl";
#    if (-e $filename) {
#        $content = loadFile($filename);
#    } else {
#        $content = "This is the $modulename page";
#    }
#
#    sendto_screen('', ucfirst("$modulename:  $noticeMsg"), $content);
#    return ($nextCommand, $nextQuery, $noticeMsg, $on);
#}

=head2 do_userpage

Display the admin page for a user\'s account. More pseudo-TT stuff.
(Userid is $defaults->{user_id})

=cut

sub do_userpage {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    my $maxitems = 10;

    my ( $content, $right_column );
    my $user_id = $defaults->{ user_id } || die "User ID number not found";

    $filter =~ s/\[\%\s*user_id\s*\%\]/$user_id/g;

    # Load up the right hand column of commands
    my $showsection = 1;
    my $title       = '';
    open( RIGHTCOLUMN, "<$srcDir/$modulename/display.right_column.tmpl" );
    while ( <RIGHTCOLUMN> ) {
        if ( $auth_user ) {
            s/\[\%\s*Username\s*\%\]/$auth_user/g;
            s/\[\%\s*$idField\s*\%\]/$user_id/g;
        } else {
            s/\[\%\s*Username\s*\%\]/Anonymous/g;
            s/\[\%\s*$idField\s*\%\]/NONE/g;
        }

        # User commands
        while (
/\[\%\s*command\s*\(\s*\"(\w+)\"\,\s*\"([\w\&\=\-]+)\"\,\s*\"([\w\s]+)\"\s*\)\s*\%\]/
          )
        {
            my $action = '';
            if ( $showsection == 1 ) {
                my $cmd = cgiCmd( $1, $2 );
                my $item = qq|<A HREF="$cmd">$3</A><BR>|;
                $action = loadFile( "$srcDir/box_row.tmpl" );
                $action =~ s/\[\%\s*item\s*\%\]/$item/;
            }
s/\[\%\s*command\s*\(\s*\"(\w+)\"\,\s*\"([\w\&\=\-]+)\"\,\s*\"([\w\s]+)\"\s*\)\s*\%\]/$action/;
        }
        while (
               /\[\%\s*section\s*\(\s*\"(\w+)\"\,\s*\"([\s\w]+)\"\s*\)\s*\%\]/ )
        {
            my $sec = '';
            $title = $2;

            # Check to see if user has permissions for this area
            if ( exists $defaults->{ Permissions }->{ $1 }
                 && $defaults->{ Permissions }->{ $1 } eq 'Y' )
            {
                $showsection = 1;
                $sec         = loadFile( "$srcDir/box_start.tmpl" );
                $sec =~ s/\[\%\s*box_title\s*\%\]/$title/g;
            } else {
                $showsection = 0;
            }
s/\[\%\s*section\s*\(\s*\"(\w+)\"\,\s*\"([\s\w]+)\"\s*\)\s*\%\]/$sec/;
        }
        while ( /\[\%\s*altsection\s*\%\]/ ) {
            my $alt = "";

            if ( $showsection == 0 ) {
                $showsection = 1;
                $alt         = loadFile( "$srcDir/box_start.tmpl" );
                $alt =~ s/\[\%\s*box_title\s*\%\]/$title/;
            } else {
                $showsection = -1;
            }

            # Check to see if user has permissions for this area
            s/\[\%\s*altsection\s*\%\]/$alt/;
        }
        while ( /\[\%\s*endsection\s*\%\]/ ) {
            my $end = "";
            if ( $showsection != 0 ) {
                $end = loadFile( "$srcDir/box_end.tmpl" );
            } else {
                $showsection = 1;
            }

            # Check to see if user has permissions for this area
            s/\[\%\s*endsection\s*\%\]/$end/;
        }
        $right_column .= $_;
    }
    close( RIGHTCOLUMN );

    open( USERPAGE, "<$srcDir/$modulename/display.single.$command.tmpl" );
    while ( <USERPAGE> ) {
        if ( $auth_user ) {
            s/\[\%\s*Username\s*\%\]/$auth_user/g;
            s/\[\%\s*$idField\s*\%\]/$user_id/g;
        } else {
            s/\[\%\s*Username\s*\%\]/Anonymous/g;
            s/\[\%\s*$idField\s*\%\]/NONE/g;
        }
        while ( /\[\%\s*eidetic\s*(\w+)\s*(\w+)\s*(\w+)\s*\%\]/ ) {
            my $htmlTable =
              makeHtmlTable( $dbh, 'stp', $1, $2, $3, $filter, $sort, $maxitems,
                             $defaults );
            s/\[\%\s*eidetic\s*(\w+)\s*(\w+)\s*(\w+)\s*\%\]/$htmlTable/;
        }
        while ( /\[\%\s*eidetic\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*\%\]/ ) {
            my $htmlTable =
              makeHtmlTable( $dbh, $1, $2, $3, $4, $filter, $sort, $maxitems,
                             $defaults );
s/\[\%\s*eidetic\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*\%\]/$htmlTable/;
        }

        # Extra-module commands
        while ( /\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/ ) {
            my $cmd = cgiCmd( 'stp', $2 );
            my $action = qq|<A HREF="$cmd">$2</A>|;
            s/\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/$action/;
        }

        # Intra-module commands
        while ( /\[\%\s*command\s+(\w+)\s*\%\]/ ) {
            my $cmd = cgiCmd( $modulename, $1 );
            my $action = qq|<A HREF="$cmd">$1</A>|;
            s/\[\%\s*command\s+(\w+)\s*\%\]/$action/;
        }
        $content .= $_;
    }
    close( USERPAGE );

    # Substitute remaining values as required
    $content =~ s/\[\%\s*noticeMsg\s*\%\]/$noticeMsg/g;
    $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
    $content =~ s/\[\%\s*modulename\s*\%\]/$modulename/g;
    if ( DEBUGGING ) {
        $content =~ s/\[\%\s*(\w+)\s*\%\]/\[\% ERROR!! $1 \%\]/g;
    } else {
        $content =~ s/\[\%\s*(\w+)\s*\%\]//g;
    }

    # More proper would be to have a content builder that included
    # provision for setting notices...
    sendto_screen( '', ucfirst( "$modulename:  $noticeMsg" ), $content, '',
                   $right_column );
    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head2 do_redirect

Generates a redirect to $dataHandle->{path} & {file}.

=cut

sub do_redirect {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    my $uri = $dataHandle->{ path } . $dataHandle->{ file };
    print $cgi->redirect( $uri );
}

=head2 do_search

Prints a search form or handles the input from said form.

=cut

sub do_search {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    if ( !%$dataHandle ) {
        my $content = loadFile( "$srcDir/$modulename/search.tmpl" );
        $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
        sendto_screen( '', "Advanced Search", $content );
        return ( $nextCommand, $nextQuery, $noticeMsg );
    }
    while ( my ( $key, $value ) = each %$dataHandle ) {

        # TODO:  Add ability to handle non-strings
        # TODO:  Add capability to specify GT, LT, EQ, CN, etc.
        if ( $value ne '*' && $value ne '' ) {
            if ( $filter ne '' ) {
                $filter .= " AND ";
            }
            $value =~ s/\%/PERCENT/g;
            $value =~ s/\*/\%/g;
            $value = '%' . $value . '%';
            $value =~ s/PERCENT/\\\%/g;

            if ( $key =~ /^(\w+)\_$idField$/ ) {
                my $source = $1;
                my $sth    =
                  selectRecords( $dbh, lc( $source ),
                                 "descriptor LIKE '$value'", '' );
                my @ids = ();
                while ( my $item = $sth->fetchrow_hashref ) {
                    my $sourceId = $item->{ $idField };
                    if ( $sourceId ) {
                        push @ids, $sourceId;
                    }
                }
                if ( @ids > 1 ) {
                    $filter .= "$key IN (" . join ( ',', @ids ) . ")";
                } elsif ( @ids == 1 ) {
                    $filter .= "$key = $ids[0]";
                } else {
                    die "The search item '$value' matches no known $source";
                }
            } elsif ( $key =~ /^\w+\_$idField\_(\w+)\.(\w+)$/ ) {
                die "Searching on x_uid_(y).(z) not implemented\n";
            } elsif ( $key =~ /^\w+\_$idField\_(\w+)$/ ) {
                my $source = $1;

                my $sth =
                  selectRecords( $dbh, lc( $source ),
                                 "descriptor LIKE '$value'", '' );
                my @ids = ();
                while ( my $item = $sth->fetchrow_hashref ) {
                    my $sourceId = $item->{ $idField };
                    if ( $sourceId ) {
                        push @ids, $sourceId;
                    }
                }
                if ( @ids > 1 ) {
                    $filter .= "$key IN (" . join ( ',', @ids ) . ")";
                } elsif ( @ids == 1 ) {
                    $filter .= "$key = $ids[0]";
                } else {
                    die "The search item '$value' matches no known $source";
                }
            } else {
                $filter .= "$key LIKE \'$value\'";
            }
        }
    }
    do_many(
             $modulename, $command,  $table,   $dbh,
             $on,         $defaults, $noticeMsg
    );

    return ( $nextCommand, $query, $noticeMsg );
}

=head2 do_display

Parses $on and $query, then sendto_screens records according to the HTML 
template. (Uses do_single and do_many)

=cut

sub do_display {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    die "Error - table name not given for $command $modulename (on $on)\n"
      unless $table;

    $defaults->{ noticeMsg } = $noticeMsg;
    if ( $query =~ /^[A-Za-z]\w+$/ ) {
        ( $table, $filter, $sort ) = getQuery( $modulename, $query );
        if ( $on eq '' ) {
            $on = "0-";
        }
        $filter =~ s/\[\%\s*user_id\s*\%\]/$defaults->{user_id}/g;
        die "I don't know what the query '$query' means<BR>\n"
          unless ( $table );
    } elsif ( $on eq '' ) {
        $on = '0-';
    }
    if ( $query eq '' ) {
        $nextQuery = 'all';
        $query     = 'all';
    }

    if ( $on =~ /^\d+$/ ) {
        do_single(
                   $modulename, $command,  $table,   $dbh,
                   $on,         $defaults, $noticeMsg
        );
    }

    if ( $on =~ /^[\-\d]\d*[\!\,\-][\,\-\d]*$/ ) {
        do_many(
                 $modulename, $command,  $table,   $dbh,
                 $on,         $defaults, $noticeMsg
        );
    }
    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head2 do_cancel_confirm

Prepares to cancel an STP test & requests confirmation

=cut

sub do_cancel_confirm {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;

    # Validate inputs
    die "Cannot use command '$command' on '$modulename'\n"
	unless ($modulename eq 'stp');
    die "Invalid test request id '$on'\n"
	unless ($on != /^\d+$/);
    die "Invalid database handle\n"
	unless ($dbh);

    # Attempt to retrieve the test request object
    my $sth = selectRecords( $dbh, "$table", "$idField=$on", "");
    die "Error retrieving test request record '$on' from database\n"
	unless ($sth);
    my $test_request = $sth->fetchrow_hashref;

    # Verify that the test request is in a cancellable state
    die "Test request id '$on' is in an undefined status\n"
	unless (defined $test_request->{status});
    die "Test request id '$on' is in state '$test_request->{status}', but only 'Queued' tests can be cancelled\n"
	unless ($test_request->{status} eq 'Queued');

    # Verify user is allowed to cancel it
    die "You do not have permission to cancel test request id '$on'\n"
	unless ($test_request->{created_by} == $defaults->{$idField}
		|| $defaults->{Permission}->{Stp_admin} eq 'Y');

    # Display cancel confirmation page
    my $content = loadFile( "$srcDir/stp/cancel_confirm.tmpl" );
    $content = substituteValues( $content, $test_request );

    sendto_screen( '', "Cancel a Test Request", $content );

    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head2 do_cancel

Cancels a requested STP test

=cut

sub do_cancel {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;
    
    # Validate inputs
    die "Cannot use command '$command' on '$modulename'\n"
	unless ($modulename eq 'stp');
    die "Invalid test request id '$on'\n"
	unless ($on != /^\d+$/);
    die "Invalid database handle\n"
	unless ($dbh);

    # Attempt to retrieve the test request object
    my $sth = selectRecords( $dbh, "$table", "$idField=$on", "");
    die "Error retrieving test request record '$on' from database\n"
	unless ($sth);
    my $test_request = $sth->fetchrow_hashref;

    # Verify that the test request is in a cancellable state
    die "Test request id '$on' is in an undefined status\n"
	unless (defined $test_request->{status});
    die "Test request id '$on' is in state '$test_request->{status}', but only 'Queued' tests can be cancelled.  Please contact the STP admins to cancel this test.\n"
	unless ( uc($test_request->{status}) eq uc('Queued')
		 || uc($test_request->{status}) eq uc('Installing OS') );

    # Verify user is allowed to cancel it
    die "You do not have permission to cancel test request id '$on'\n"
	unless ($test_request->{created_by} == $defaults->{$idField}
		|| $defaults->{Permission}->{linux_stp_admin} eq 'Y');

    ### Update record and mark test request cancelled
    my $sql = qq|UPDATE test_request SET status='Cancelled' WHERE $idField=$on|;
    $sth = $dbh->prepare($sql) 
	or die "Could not prepare database command: " . $dbh->errstr;
    my $rows_affected = $sth->execute();

    die "Error trying to set test request to status 'Cancelled': " . $dbh->errstr . "\n"
	if ($rows_affected == 0);

    return ( 'userpage', '', 'Test Request Has Been Cancelled' );
}

=head2 do_edit

Displays, via HTML, the edit form for $modulename and $query. Uses do_single and
do_many.

=cut

sub do_edit {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;

    if ( $query ne '' ) {
        ( $table, $filter, $sort ) = getQuery( $modulename, $query );
    }

    # This needs to be refactored to allow the form specification to
    # be more generalized and abstracted...  I need to take into
    # consideration how the admin might want to edit these forms...
    if ( ( $on eq '' && $query eq '' )
         || ( $on =~ /^\d+$/ && ( $modulename eq 'tests' ) )
         || ( $on =~ /^\d+$/ && ( $modulename eq 'hosts' ) )
         || ( $on =~ /^\d+$/ && ( $modulename eq 'host_types' ) )
         || ( $on eq '' && $modulename eq 'articles' )
         || ( $on eq '' && $modulename eq 'documents' ) )
    {
        my $fname = "$htmlDir/$modulename/$table.html";
        my $sname = "$srcDir/$modulename/edit.$table.$query.form.tmpl";
        if ( !-f $sname ) {
            $sname = "$srcDir/$modulename/edit.$table.form.tmpl";
        }

        my $content    = '';
        my $hiddenInfo = '';

        if ( $on =~ /^\d+$/ ) {
            if ( $modulename eq 'users' && $defaults->{ $idField } ne $on ) {
                if ( $defaults->{ Permissions }->{ admin } ne 'Y' ) {
                    die "Cannot access record for $on\n";
                }
            }

            my $id =
              $on;   # need to incorporate numbering a little more appropriately
            if ( !isInDatabase( $dbh, $table, "$idField=$id" ) ) {
                die "Couldn't find $id\n";
            }
            my $sth = selectRecords(
                                     $dbh,           $table,
                                     "$idField=$id", '',
                                     $start,         $maxitems
            );
            if ( $sth == 0 ) {

                # No available records were found
                # TODO:  Tell user we couldn't process it
                die "Couldn't process $id!\n";
            }
            $data = $sth->fetchrow_hashref;
        }

        if ( $modulename eq 'stp' ) {
            if ( $command eq 'apply' ) {
                $sname = "$srcDir/$modulename/edit.apply.form.tmpl";
            } elsif ( !exists $data->{ test_uid } ) {
                $sname = "$srcDir/$modulename/edit.$table.info.form.tmpl";
            } else {

                # On to page #2
                foreach my $key ( keys %$data ) {
                    next if $key eq 'patch_tag';
                    next if $key eq 'distro_tag_uid';
                    next if $key eq 'lilo';
                    next if $key eq 'host_type_uid';
                    $hiddenInfo .= qq|   <INPUT TYPE="hidden" NAME="d_|;
                    $hiddenInfo .= $key;
                    $hiddenInfo .= qq|" VALUE="|;
                    $hiddenInfo .= $data->{ $key };
                    $hiddenInfo .= qq|">\n|;
                }

                # Look up the default lilo string for test_uid

                my $sth =
                  selectRecords( $dbh, 'test',
                                 "$idField=" . $data->{ test_uid }, '', $start,
                                 $maxitems );
                if ( $sth != 0 ) {
                    my $row = $sth->fetchrow_hashref;
                    $data->{ lilo } = $row->{ default_lilo } || '';
                } else {
                    $data->{ lilo } = 'Could not find test';
                }

                $sname =
                  "$srcDir/$modulename/edit.$table.machine_setup.form.tmpl";
            }
        }

        if ( -e $sname ) {

            # Substitute any dropdown boxes with dynamic data
            $content = loadFile( $sname );
            $content =~ s/\[\%\s*Hidden_info\s*\%\]/$hiddenInfo/;
            $content =~ s/\[\%\s*Notice\s*\%\]/$noticeMsg/;
            $content =~ s/\[\%\s*noticeMsg\s*\%\]/$noticeMsg/;
            $content =~ s/\[\%\s*user_id\s*\%\]/$defaults->{user_id}/g;
            $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
            $content =~
              s/\[\%\s*Reqd\s*\%\]/\<FONT COLOR\=\'red\'\>\*\<\/FONT\> /g;
            $content = substituteValues( $content, \%storageError );
            $content = substituteValues( $content, $data );

            # I don't think we actually want to substitute in user info...
#            $content = substituteValues($content, $defaults, '&nbsp;');

            # Some user info...
            if ( $defaults->{ Real_email } ) {
                $content =~
                  s/\[\%\s*Real_email\s*\%\]/$defaults->{Real_email}/g;
            }
            if ( $defaults->{ descriptor } ) {
                $content =~ s/\[\%\s*Username\s*\%\]/$defaults->{descriptor}/g;
            }

            my $id = $data->{ $idField };

            my $table = '';
            if ( $modulename eq 'host_types' ) {
                $table = 'host_type_temp';
            } elsif ( $modulename eq 'tests' ) {
                $table = 'test';
            } elsif ( $modulename eq 'hosts' ) {
                $table = 'host_temp';
            } elsif ( $modulename eq 'projects' ) {
                $table = 'project';
            } elsif ( $modulename eq 'users' or $modulename eq 'persons' ) {
                $table = $person_table;
            }
            while ( $content =~ /\[\%\s*checkbox-list\s*(\w+)\s*(\w+)\s*\%\]/ )
            {
                my ( $junction, $target, $source ) = ( $1, $2, $table );

                my $sth;
                if ( $on ne "" ) {
                    $sth =
                      selectRecords( $dbh, "$target<-+$junction",
                                     "$junction.$source\_$idField=$id",
                                     '' );
                } else {

                    # TODO:  uninitialized value in string error
                    $sth = selectRecords( $dbh, "$target", '', '' );
                }
                my $text = '';
                while ( my $item = $sth->fetchrow_hashref ) {
                    my $i = $item->{ $idField } || '';
                    $text .= qq|<INPUT TYPE="checkbox" VALUE="$i" |;

                    # TODO:  uninitialized value in string error
                    if ( $item->{ "$source\_$idField" } ne '' ) {
                        $text .= "CHECKED ";
                    }
                    $text .=
qq|NAME="d_$junction\:\:$target\_$idField\[$i\]"> $item->{descriptor}<BR>\n|;
                }
                $content =~
                  s/\[\%\s*checkbox-list\s*$junction\s*$target\s*\%\]/$text/;
            }
            $content = processPage( $content, $data );
            $content =~ s/\[\%.*\%\]//g;

            sendto_screen( '', ucfirst( "$modulename:  Submit $table" ),
                           $content );
        } elsif ( -e $fname ) {
            my $content = loadFile( $fname );

            print $content;
        } else {

            # Display a blank edit form for this
# If a template was specified, use it, otherwise generate the default
            if ( $on =~ /^\d+$/ ) {
                $data = getTableInfo( $dbh, $table, 'DEFAULTS' );
            }

            my %form = (
                         fields => getTableInfo(
                                                 $dbh, $table,
                                                 'INPUT_FIELDS'
                         ),
                         descriptions => getTableInfo(
                                                       $dbh, $table,
                                                       'DESCS'
                         ),
                         types  => getTableInfo( $dbh, $table, 'TYPES' ),
                         values => $data,
                         notes  => getTableInfo( $dbh, $table, 'EXTRA' )
            );
            $content = genericHtmlForm( $modulename, 'store', %form );

            $table =~ s/\_/ /g;
            $table =~ s/s$//;

            sendto_screen( '', ucfirst( "$modulename:  Submit $table" ),
                           $content );
        }
    }
    if ( $on =~ /^\d+$/
         && $modulename ne 'tests'
         && $modulename ne 'hosts'
         && $modulename ne 'host_types' )
    {

        do_single(
                   $modulename, $command,  $table,   $dbh,
                   $on,         $defaults, $noticeMsg
        );
    }
    if ( $on =~ /^[\-\d]\d*[\!\,\-][\,\-\d]*$/ ) {
        do_many(
                 $modulename, $command,  $table,   $dbh,
                 $on,         $defaults, $noticeMsg
        );
    } else {

        # User entered strange value for $on; need to ask for clarification...
    }
    if ( $query eq 'all' ) { $query = ''; $nextQuery = $query; }

    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head2 validate(VALUE, FORMATTAG)

Validates that $value is a valid $formatTag; returns list with 2 empty elements
if so, list with 'reedit' and an error message otherwise.

=cut

sub validate($$) {
    my ( $value, $formatTag ) = @_;
    my $errorMsg;

    # Eventually, this will load its regexp's from a database table

    if ( $formatTag eq 'name' && $value !~ /^[\w\.\,\-\' ]{1,80}$/ ) {
        $errorMsg =
"This must be between 1 and 80 characters in length and use only the characters 0-9, a-z, A-Z, and the following punctuation:  _.,-'";
    } elsif ( $formatTag eq 'email' && $value !~ /^[\w\.\@\-]{5,255}$/ ) {
        $errorMsg =
"This must be between 5 and 255 characters in length and use only the characters 0-9, a-z, A-Z, and the following punctuation: _.@-";
    } elsif ( $formatTag eq 'username' && $value !~ /^[a-z0-9\_]{2,20}$/ ) {
        $errorMsg =
"This must be between 2 and 20 characters in length and use only the characters 0-9, a-z, and _.";
    } else {
        return ( '', '' );
    }
    return ( 'reedit', $errorMsg );
}

=head2 do_storefile

Saves $data->{content} as $data->{filename}, and runs TT2 on it.

=cut

sub do_storefile {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;

    if ( $data->{ Filename } !~ /^\w[\w\/\.]+/ ) {

        # TODO - Add error handling
        die ( "Unacceptable characters in filename" );
    }

    $nextCommand = 'display';
    my $ext = "tmpl";
    if ( $data->{ Language } eq 'Japanese' ) {
        $ext = "jp.tmpl";
    }
    my $filename = "$srcDir/$data->{Filename}.$ext";
    my $target   = "$htmlDir/$data->{Filename}.html";
    my @path     = split /\//, $filename;
    pop @path;
    my $path = join '/', @path;
    my $content = $data->{ Content };
    if ( -e $filename || -d $path ) {
        sendto_file( $filename, '', $content );
        `cd $path; /usr/bin/tpage $filename > $target`;
        $noticeMsg = "Stored $filename";
    } else {
        die "Could not create file";
    }
    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head2 do_store

Stores data into the given module\'s table. If $on is empty, adds a new record;
otherwise updates records given by $on. Special cases for users, stp, and 
projects.

=cut

sub do_store {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    my $content = '';
    $nextCommand = 'display';
    my $nextOn  = '';
    my $user_id = $defaults->{ user_id } || 0;

    if ( $on eq '' ) {
        my $message = '';
        my @next    = ();
        %storageError = ();

        # Special code for users module
        if ( $modulename ne 'stp'
            && $modulename ne 'articles'
            && $modulename ne 'documents'
            && isInDatabase(
                  $dbh, $table,
                  "descriptor='" . $dataHandle->{ descriptor } . "'"
            ) )
        {
            $content .= $dataHandle->{ descriptor } . ' already exists.';
            $storageError{ Descriptor_Error } =
              'This is already used, try another.';
            $storageError{ Username_Error } =
              'This is already used, try another.';
            $next[ 0 ] = 'reedit';
        } elsif ( $modulename eq 'users' ) {

            # Validate the user's info
            ( $next[ 0 ], $storageError{ Given_Name_Error } ) =
              validate( $dataHandle->{ Given_name }, 'name' );
            ( $next[ 1 ], $storageError{ Surname_Error } ) =
              validate( $dataHandle->{ Surname }, 'name' );
            ( $next[ 2 ], $storageError{ Real_Email_Error } ) =
              validate( $dataHandle->{ Real_email }, 'email' );
            if ( $next[ 2 ] eq '' && isInDatabase(
                          $dbh,
                          $table,
                          "Real_email='" . $dataHandle->{ Real_email } . "'"
                ) )
            {
                $next[ 2 ] = 'reedit';
                $storageError{ Real_Email_Error } =
                  "This email address is already registered.";
            }

            ( $next[ 3 ], $storageError{ Username_Error } ) =
              validate( $dataHandle->{ descriptor }, 'username' );
            $dataHandle->{ Username } = $dataHandle->{ descriptor };

            my $sth = $dbh->prepare( "SELECT NOW()" );
            $sth->execute;
            my $now = $sth->fetchrow;

            $dataHandle->{ Join_date } = $now;
        } elsif ( exists $dataHandle->{ Apply_for_role } ) {
            my $role = $dataHandle->{ Apply_for_role };
            if ( !exists $dataHandle->{ Accept_conditions }
                 || $dataHandle->{ Accept_conditions } ne 'Y' )
            {
                $storageError{ Accept_Conditions_Error } =
                  "If you do not accept the conditions, "
                  . "we cannot provide access.";
                $next[ 3 ] = 'apply';
            } else {
                $noticeMsg = "You now have permission to use STP.";
                changePermission( $dbh, $user_id, 'stp_user', 'Y' );
            }
        }

        my $new_file = '';
        if ( join ( '', @next ) ne '' ) {

            # Break here and require user to reedit
            $noticeMsg = "Please check your entry and re-edit...";
            return ( 'reedit', $query, $noticeMsg );
        } elsif ( $modulename eq 'users' ) {

            # Generate a new password
            my $passwd = generateRandomPassword();

            # Prepare email msg for user
            $dataHandle->{ Userpassword } = $passwd;
            $message = loadFile( "$srcDir/users/new.user.email.tmpl" );
            $message = substituteValues( $message, $dataHandle );
            if ( DEBUGGING ) {
                $message =~ s/\[\%\s*(\w+)\s*\%\]/\[\% ERROR!! $1 \%\]/g;
            } else {
                $message =~ s/\[\%\s*(\w+)\s*\%\]//g;
            }

            # Encrypt password
            my $salt = sprintf( "\$1\$%.8s\$", generateRandomPassword() );
            $dataHandle->{ Userpassword } = md5_hex( $passwd );
            $dataHandle->{ Unixpassword } = crypt( $passwd, $salt );

            # We're going to forward them to a thankyou page next.
            $nextCommand = 'thanks';
        } elsif ( $modulename eq 'stp' ) {

            # This is a quick kludge to implement multipage forms
            # More properly, the validation system should determine
            # if additional forms are needed and figure out what
            # they should be.
            if ( exists $dataHandle->{ patch_tag } ) {
                $nextCommand = 'userpage';

                #		if (!isInDatabase($dbh, 'patch_tag', 
                #				  "uid='".$dataHandle->{patch_tag}."'")) {
                #		    $noticeMsg = qq|
#The indicated Patch ID "$dataHandle->{patch_tag}" could not be found.
                #|;
                #		    $storageError{Patch_Tag_Error} = "Invalid patch ID";
                #               }
                if ( $dataHandle->{ patch_tag } !~ /^\d+$/ ) {
                    $noticeMsg = qq|Invalid patch tag - integers only please|;
                    $storageError{ Patch_Tag_Error } = "Invalid patch ID";
                    return ( 'reedit', $query, $noticeMsg );
                }
            } else {
                $noticeMsg = '';
                return ( 'edit', $query, $noticeMsg );
            }
        } elsif ( $modulename eq 'documents' ) {
            if ( !$dataHandle->{ Filename } ) {
                $noticeMsg = 'Error in uploading file';
                $storageError{ Filename_Error } =
                  "Filename not specified - cannot perform upload.";
                $nextCommand = 'reedit';
                return ( $nextCommand, $query, $noticeMsg );
            } else {
                $dataHandle->{ Submitter_uid_user }  = $user_id;
                $dataHandle->{ Contact_uid_user }    = $user_id;
                $dataHandle->{ Maintainer_uid_user } = $user_id;
                $dataHandle->{ Creator_uid_user }    = $user_id;
                $dataHandle->{ Author }              = $defaults->{ descriptor }
                  || 'Anonymous';
                $dataHandle->{ Status }  = 'Released';
                $dataHandle->{ Version } = 1;

                if ( !$dataHandle->{ Language_uid } ) {
                    $dataHandle->{ Language_uid } = 1001;
                }

                prepare_archive( $dataHandle );

# Temporary solution to duplicate filenames; a better solution will be needed
                if ( isInDatabase(
                              $dbh,
                              'document',
                              "Filename='" . $dataHandle->{ Filename } . "'"
                    ) )
                {
                    die
"Sorry, that title/filename conflicts with documents already in "
                      . "the system.  Click the back button and try a different name.\n";
                }

            }
        }

        while ( my ( $key, $value ) = each %$defaults ) {
            if ( !$dataHandle->{ $key }
                 && $key ne 'descriptor'
                 && $key ne $idField
                 && $key ne 'rsf' )
            {
                $dataHandle->{ $key } = $value;
            }
        }

        # Validate the inputs and store the data
        my %data = validateAndStoreRecord(
                                           $modulename, $user_id,
                                           %$dataHandle
        );

        if ( $modulename eq 'documents' && defined $data{ Language_uid } ) {
            my $sth =
              selectRecords( $dbh, 'language', "uid=$data{Language_uid}", '' );
            if ( $sth != 0 ) {
                $data{ Language } = ( $sth->fetchrow_hashref )->{ Charset };
            }
        }

        # Perform file upload processing
        if ( ( $modulename eq 'documents' ) && $data{ Filename } ne '' ) {
            my $destination = $dataHandle->{ Submitter_uid_user } || 'incoming';
            my $archive =
              receive_archive(
                               $data{ Filename }, $data{ $idField },
                               $data{ Language }, $data{ Version },
                               "$uploadDir/$destination"
            );
            if ( !$archive ) {
                die "The file could not be entered into the repository.\n";
            }
            my $errorMsg =
              process_archive(
                               $archive, $data{ Filename },
                               ( "$htmlDir/docs" )
            );
            die "Error processing $errorMsg\n" unless ( !$errorMsg );

            $noticeMsg = qq|
                <P><B>'$data{Filename}'</B> has been successfully received 
                into the repository as document # $data{$idField}, 
                ver. $data{Version}.  You can access it from the page of
                the <A HREF="$my_url?command=display&modulename=projects&on=$data{Project_uid}">project owning it</A>, 
                or directly using <A HREF="/docs/$data{Filename}">this link</A>.

                <P>If you suspect any problems, please email 
                <A HREF="mailto:$webmaster">$webmaster</A>.</P>
                |;
        }

        if ( $modulename eq 'users' ) {
            sendto_email( $dataHandle->{ Real_email }, "User Account",
                          $message );
        } elsif ( $modulename eq 'projects' ) {
            $noticeMsg =
qq|Thank you for your project request, it will be reviewed and you will be emailed notification of its status.|;
            if ( `hostname` !~ /^osdlab/ ) {
                notifyStaff( $modulename, %data );
            }
        }

# if they've requested to be entered into the mailing list, then add them
        if ( $data{ Maillist } eq "Y" ) {
            my ( $fh, $fname );
            do { $fname = tmpnam(); }
              until $fh = IO::File->new( $fname, O_RDWR | O_CREAT | O_EXCL );

            END {
                if ( $fname ) {
                    unlink( $fname )
                      or die "failure Couldn't unlink $fname : $!";
                }
            }

            sendto_file( $fname, '', join ( "\n", $data{ Real_email } ) );

            # TODO:  Path should be a config variable
`/var/mailman/bin/add_members --non-digest-members-file=$fname osdl_associates`;
            $content .=
              "You've been added to the osdl_associates mailing list.\n";
        }

        # Fall through to display routine down below.
        if ( $noticeMsg eq '' ) {
            $noticeMsg = "Record $data{$idField} was stored successfully.  ";
        }
        $nextOn = $data{ $idField };
    }
    if ( $on =~ /^\d+$/ ) {
        if ( $modulename eq 'users' && $defaults->{ $idField } ne $on ) {
            if ( $defaults->{ Permissions }->{ admin } ne 'Y' ) {
                die
"Cannot access record for $on - must have admin permissions\n";
            }
        }

        # Store edits to a particular record

        my $owner = 1;

        #       authorizeUser($dbh, $auth_user, "uid='$owner'") ||
        #           die "You do not have access to change this password.";

        if ( $modulename eq 'documents' ) {

            # Get the current version number & verify this user can do editing
            my $old_filename = '';
            my $sth = selectRecords( $dbh, 'document', "$idField=$on", '' );
            if ( $sth != 0 ) {
                my $old_version = $sth->fetchrow_hashref;

                if ( $old_version->{ Submitter_uid_user } != $user_id ) {
                    my $owner =
                      selectRecords( $dbh, $person_table,
                                  'uid=' . $old_version->{ Submitter_uid_user },
                                  '' );
                    my $owner_info = '';
                    if ( $owner != 0 ) {
                        $owner_info =
                          '. The owner is '
                          . ( $owner->fetchrow_hashref )->{ descriptor };
                    }
                    die
"Only the owner of this document can update it $owner_info";
                }
                $dataHandle->{ Version } = $old_version->{ Version } + 1;
                $old_filename = $old_version->{ Filename };
            }

            $dataHandle->{ Submitter_uid_user } = $user_id;

            prepare_archive( $dataHandle );

# Temporary solution to duplicate filenames; a better solution will be needed
            if ( $dataHandle->{ Filename } ne $old_filename && isInDatabase(
                              $dbh,
                              'document',
                              "Filename='" . $dataHandle->{ Filename } . "'"
                ) )
            {
                die "Sorry, that title/filename ("
                  . ( $dataHandle->{ Filename } )
                  . ") conflicts with documents already in the system. "
                  . " Click the back button and try a different name.\n";
            }

        }

        # Validate the inputs and store the data
        my %data = updateRecord( $modulename, $user_id, %$dataHandle );

        if ( $modulename eq 'documents' && defined $data{ Language_uid } ) {
            my $sth =
              selectRecords( $dbh, 'language', "uid=$data{Language_uid}", '' );
            if ( $sth != 0 ) {
                $data{ Language } = ( $sth->fetchrow_hashref )->{ Charset };
            }
        }

        # Perform file upload processing
        if ( ( $modulename eq 'documents' || $modulename eq 'articles' )
             && $data{ Filename } )
        {

            # Uninstall any old files from whereever they're installed
            my $sth =
              selectRecords( $dbh, 'document', "uid=$data{$idField}", '' );
            if ( $sth != 0 ) {
                clean_archive( ( $sth->fetchrow_hashref )->{ Filename },
                               ( "$htmlDir/docs" ) );

            }

            my $destination = $dataHandle->{ Submitter_uid_user } || 'incoming';
            if ( !-e "$uploadDir/$destination" ) {
                print `mkdir $uploadDir/$destination`;
            }
            my $archive =
              receive_archive(
                               $data{ Filename }, $data{ $idField },
                               $data{ Language }, $data{ Version },
                               "$uploadDir/$destination"
            );
            if ( !$archive ) {
                die "The file could not be entered into the repository.\n";
            }

            my $errorMsg =
              process_archive(
                               $archive, $data{ Filename },
                               ( "$htmlDir/docs" )
            );
            die "Error processing $errorMsg\n" unless ( !$errorMsg );

            $noticeMsg = qq|
                <P><B>'$data{Filename}'</B> has been successfully received 
                into the repository as document # $data{$idField}, 
                ver. $data{Version}.  You can access it from the page of
                the <A HREF="$my_url?command=display&modulename=projects&on=$data{Project_uid}">project owning it</A>, 
                or directly using <A HREF="/docs/$data{Filename}">this link</A>.

                <P>If you suspect any problems, please email 
                <A HREF="mailto:$webmaster">$webmaster</A>.</P>
                |;
        }

        # Fall through to display routine down below.
        if ( $noticeMsg eq '' ) {
            $noticeMsg = "Record $data{$idField} was updated successfully.  ";
        }
        $nextOn = $on;
    }
    if ( $on =~ /^\d+\-\d+$/ || $on =~ /^\d+\-$/ ) {
        if ( $modulename eq 'users' && $defaults->{ $idField } ne $on ) {
            if ( $defaults->{ Permissions }->{ admin } ne 'Y' ) {
                die
"Cannot store records for $on - must have admin permissions\n";
            }
        }

        # Storing a range of records
        my %records;

        # Split the data into an array of records indexed by id
        while ( my ( $key, $value ) = each %$dataHandle ) {
            if ( $key =~ /$idField\((\d+)\).(\w+)/ ) {
                $records{ $1 }->{ $2 } = $value;
            }
        }

        foreach my $rec_id ( keys %records ) {
            $records{ $rec_id }->{ $idField } = $rec_id;
            my %rec = updateRecord(
                                    $modulename, $user_id,
                                    %{ $records{ $rec_id } }
            );
        }
        if ( $modulename eq 'projects' ) {
            sendto_email( $proj_update_email, "Project Status Updates",
                          "The following project ID's have been updated:  "
                          . join ( ', ', keys %records ) );
        }

        # Fall through to display routine down below.
        if ( $noticeMsg eq '' ) {
            $noticeMsg = "Records $on updated.";
        }
    }
    return ( $nextCommand, $nextQuery, $noticeMsg, $nextOn );
}

sub prepare_archive {
    my ( $dataHandle ) = @_;

    my $from_filename = $dataHandle->{ Filename };
    my $ext           = 'txt';
    my $root          = $from_filename;
    if ( $from_filename =~ /^(.*)\.(\w+)$/ ) {
        $root = $1;
        $ext  = $2;
    }
    if ( $dataHandle->{ Title } eq '' ) {
        $dataHandle->{ Title } = $root;
    }
    $dataHandle->{ Title } =~ s/[\/\\]/\_/g;
    my $filename = lc( cleanFilename( $dataHandle->{ Title } ) ) . '.' . $ext;
    $dataHandle->{ Filename }  = $filename;
    $dataHandle->{ Extension } = $ext;

    # Store file to a temporary location
    my ( $bytes, $buffer, $bytesread );
    if ( !-d "$tmpDir" ) { print `mkdir $tmpDir`; }
    if ( !open( UPLOADFILE, ">$tmpDir/$filename" ) ) {
        die "There was an error storing $from_filename to $tmpDir/$filename\n";
    }
    while ( $bytes = read( $from_filename, $buffer, 1024 ) ) {
        $bytesread += $bytes;
        print UPLOADFILE $buffer;
    }
    close( UPLOADFILE );

    # Get the archive's mimetype
    $dataHandle->{ Mimetype } = `file -bi $tmpDir/$filename`;
    chomp $dataHandle->{ Mimetype };

    # Get the size of the archive
    my $inode = stat( "$tmpDir/$filename" );
    if ( $inode->size < 1024 ) {
        $dataHandle->{ Filesize } = $inode->size;
    } elsif ( $inode->size < 1024 * 1024 ) {
        $dataHandle->{ Filesize } =
          sprintf( "%.1fk", ( $inode->size / ( 1024 ) ) );
    } elsif ( $inode->size < 1024 * 1024 ) {
        $dataHandle->{ Filesize } =
          sprintf( "%.1fM", ( $inode->size / ( 1024 * 1024 ) ) );
    } elsif ( $inode->size < 1024 * 1024 ) {
        $dataHandle->{ Filesize } =
          sprintf( "%.1fG", ( $inode->size / ( 1024 * 1024 * 1024 ) ) );
    }

    return catfile( $tmpDir, $filename );
}

# Takes a file from the tmpDir and moves it into the repository
# We assume the archive has already been validated
sub receive_archive {
    my ( $filename, $id, $ln, $rev, $uploadDir ) = @_;
    die "Filename not known\n"   unless ( $filename );
    die "ID not known\n"         unless ( $id );
    die "Language not known\n"   unless ( $ln );
    die "Revision not known\n"   unless ( $rev );
    die "Upload dir not known\n" unless ( $uploadDir );

    my $origin = "$tmpDir/$filename";
    die "Uploaded file $origin does not exist\n" unless ( -e $origin );

    # Snag extension from filename
    my $ext = '';
    if ( $filename =~ /\.(\w+)$/ ) {
        $ext = $1;
    }

    # Put rev into three digit form
    $rev = sprintf( "%03d", $rev );

    if ( !-e $uploadDir ) { print `mkdir $uploadDir`; }

    my $archive_file = "$id.$ext,$ln,$rev";

    # Install the file in the repository
    move( $origin, catfile( $uploadDir, $archive_file ) )
      || die "Could not move file into repository.\n";

    # Note:  The file also needs to be pulled back via scp to the
    # integration machine, but maybe that should be done via rsync 
    # externally
    return catfile( $uploadDir, $archive_file );
}

# This routine creates a thumbnail image for the given document,
# by lookup the appropriate function for its type and running it
sub create_thumbnail {

    # Disabled & removed for STP
    return -1;
}

# This routine copies the source file to appropriate external
# locations. 
sub process_archive {
    my ( $archive, $filename, @destinations ) = @_;
    my $noimage = "noimage.png";

    # Create thumbnails, if appropriate
    my $thumb = "/tmp/$filename.thumb.png";
    if ( !create_thumbnail( $archive, $thumb ) ) {

        # Could not generate a thumbnail, so use noimage.png
        $thumb = '';
    }

    my $errorMsg = '';
    foreach my $dest ( @destinations ) {
        if ( !-d $dest ) {
            print `mkdir $dest`;
            print `chmod a+rx $dest`;
        }

        if ( !-d $dest ) {
            $errorMsg .= " $dest could not be created";
            next;
        }

        if ( !copy( $archive, catfile( $dest, $filename ) ) ) {
            $errorMsg .= " $dest/$filename ";
        }

        if ( $thumb eq '' ) {
            copy( catfile( $dest, "noimage.png" ),
                  catfile( $dest, "$filename.thumb.png" ) );
        } elsif ( !copy( $thumb, "$dest/" ) ) {
            $errorMsg .= " $thumb ";
        }
    }
    return $errorMsg;
}

# Removes files associated with the previous version
sub clean_archive {
    my ( $filename, @locations ) = @_;

    my $errorMsg = '';
    foreach my $loc ( @locations ) {
        next unless ( -d $loc );

        if ( $filename =~ /^([\w\.\-]*)\.(\w+)$/ ) {

            # Remove any existing files with alternate file types
            my $basename = $1;
            my @files    = <$loc/$basename.*>;
            unlink( @files );
        }
    }
}

=head2 makeWidget(TYPE, DATA, TABLE, FIELD, DEFAULT, FIELD_LIST)

Makes various HTML widgets

=cut

sub makeWidget($$$$$;$) {
    my ( $type, $data, $table, $field, $default, $field_list ) = @_;
    my $text   = '';
    my $dbname = "eidetic";
    my $dbh    = getDbHandle( $dbname );
    my $sth;

    if ( !$table || !$field ) {

        # Bad info...
        return ( DEBUGGING ? 'Insufficient parameters' : '' );
    } elsif ( checkTableExistance( $dbh, $table ) ) {
        my @fields;
        my $source = '';
        my $target = '';
        if ( $type eq 'dropdown' ) {
            $text = "<SELECT NAME='$field'>\n";
        } elsif ( $type eq 'radio-table' ) {
            @fields = split /\|/, $field_list;
        }

        # Handle junction tables
        if ( $table =~ /^(\w+)_to_(\w+)_junction/
             || $table =~ /^(\w+)_to_(\w+)/ )
        {

            # First see if both tables actually exist
            # Next check to see what we're going from and to
            my ( $table1, $table2 ) = ( $1, $2 );
            if ( $field =~ /^d_$table1\_uid$/i ) {
                $target = $table1;
                $source = $table2;
            } elsif ( $field =~ /^d_$table2\_uid$/i ) {
                $source = $table1;
                $target = $table2;
            }

            # Making this less case sensitive...
            my $f = "$source\_$idField";
            if ( exists $data->{ ucfirst( $f ) } ) { $f = ucfirst( $f ); }

            my $id = $data->{ $f };
            $sth =
              selectRecords( $dbh, "$target<-$table",
                             "$table.$source\_$idField=$id", '' );
        } else {
            $sth = selectRecords( $dbh, $table, 'rsf=1', 'descriptor' );
        }

        while ( my $data = $sth->fetchrow_hashref ) {
            my $id = $data->{ $idField } || '';
            my $desc = '';
            if ( !$default ) {
                $default = $id;
            }
            if ( $type eq 'radio-list' ) {
                $desc = $data->{ Description } || '';
            } elsif ( $type eq 'radio-table' ) {
                $text .= " <TR><TD>";
                $desc = "</TD>\n";
                foreach my $field ( @fields ) {
                    if ( exists $data->{ lc $field } ) {
                        $field = lc $field;
                    }
                    if ( $data->{ $field } eq '' ) {
                        $desc .= qq|<TD VALIGN="top">$field</TD>|;
                    } else {
                        $desc .= qq|<TD VALIGN="top">$data->{$field}</TD>|;
                    }
                }
                $desc .= " </TR>\n";
            } else {
                $desc = $data->{ descriptor } || '';
            }
            if ( $type =~ /^radio/ ) {
                $text .= "   <INPUT TYPE='radio' NAME='$field' VALUE='$id'";
                if ( $id == $default ) { $text .= " CHECKED"; }
                $text .= "> $desc \n";
                if ( $type eq 'radio-list' ) {
                    my $cmd;
                    if ( exists $modules{ $table . 's' } ) {
                        $cmd = cgiCmd(
                                       $table . 's', 'display',
                                       ( on => $id )
                        );
                    } else {
                        $cmd = cgiCmd( $table, 'display', ( on => $id ) );
                    }

#                   $text .= qq| <A HREF="$cmd">Details...</A><BR>\n|;
                    $text .= qq| <BR>\n|;
                }
            } elsif ( $type eq 'dropdown' ) {
                $text .= "  <OPTION VALUE='$id'";
                if ( $id == $default ) { $text .= " SELECTED"; }
                $text .= "> $desc </OPTION>\n";
            }
        }
        if ( $type eq 'dropdown' ) {
            $text .= "</SELECT>\n";
            my $cmd =
              cgiCmd( $table . 's', 'display',
                      ( on => '', sort => 'descriptor' ) );
            $text .= qq|&nbsp; <A HREF="$cmd">Details...</A><BR>\n|;
        }
    } elsif ( DEBUGGING ) {

        # table wasn't found...  Hmm...
        $text = "Error: table '$table' doesn't exist";
    }
    return $text;
}

=head2 processPage(TEXT, DATA)

Does TT2-style substitutions of HTML widgets and modular commands into $text,
and returns substituted version.

=cut

sub processPage($$) {
    my ( $text, $data ) = @_;
    my $content = '';

    foreach ( split ( /\n/, $text ) ) {
        while (
      /\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/
          )
        {
            my $dropdown = makeWidget( 'dropdown', $data, $1, $2, $3 );
s/\[\%\s*dropdown\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$dropdown/;
        }
        while (
         /\[\%\s*radio\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/
          )
        {
            my $radio = makeWidget( 'radio', $data, $1, $2, $3 );
s/\[\%\s*radio\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$radio/;
        }
        while (
    /\[\%\s*radio-list\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/
          )
        {
            my $radio = makeWidget( 'radio-list', $data, $1, $2, $3 );
s/\[\%\s*radio-list\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w+)\'\s*\)\s*\%\]/$radio/;
        }
        while (
/\[\%\s*radio-table\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w*)\'\,\s*\'([\w\|]+)\'\s*\)\s*\%\]/
          )
        {
            my $radio = makeWidget( 'radio-table', $data, $1, $2, $3, $4 );
s/\[\%\s*radio-table\s*\(\s*\'(\w+)\'\,\s*\'(\w+)\'\,\s*\'(\w*)\'\,\s*\'([\w\|]+)\'\s*\)\s*\%\]/$radio/;
        }

        # Extra-module commands
        while ( /\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/ ) {
            my $cmd = cgiCmd( 'stp', $2 );
            my $action = qq|<A HREF="$cmd">$2</A>|;
            s/\[\%\s*command\s+(\w+)\:\:(\w+)\s*\%\]/$action/;
        }

        # Intra-module commands
        while ( /\[\%\s*command\s+(\w+)\s*\%\]/ ) {
            my $cmd = cgiCmd( $modulename, $1 );
            my $action = qq|<A HREF="$cmd">$1</A>|;
            s/\[\%\s*command\s+(\w+)\s*\%\]/$action/;
        }
        $content .= "$_\n";
    }
    return $content;
}

=head2 do_editfile

Displays an HTML form for editing a file.

=cut

sub do_editfile {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;

    my $ext = '.tmpl';
    if ( $data->{ Language } && $data->{ Language } eq 'Japanese' ) {
        $ext = '.jp.tmpl';
    }
    my $filename = $data->{ Filename };
    if ( !$filename ) {
        my $id =
          $on;    # need to incorporate numbering a little more appropriately
        if ( !isInDatabase( $dbh, $table, "$idField=$id" ) ) {
            die "Couldn't find $id\n";
        }
        my $sth = selectRecords( $dbh, $table, "$idField=$id", '' );
        if ( $sth == 0 ) {
            die "Could not determine filename to edit!\n";
        }
        my $metadata = $sth->fetchrow_hashref;
        $filename = $metadata->{ Filename }
          || sendto_screen( "", "No file is associated with this entry.",
                        "This $table doesn't have a file associated with it." );
    }
    my @path = split /\//, $filename;
    pop @path;
    my $path = join '/', @path;
    my $content = '';
    if ( -e "$srcDir/$filename.tmpl" ) {
        $content = loadFile( "$srcDir/$filename.tmpl" );
    } elsif ( -e "$srcDir/$path/default.tmpl" ) {
        $content = loadFile( "$srcDir/$path/default.tmpl" );
    } elsif ( !-e "$srcDir/$path/" ) {
        die "$path is not a valid place for storing files";
    } elsif ( -e "$srcDir/default.tmpl" ) {
        $content = loadFile( "$srcDir/default.tmpl" );
    } else {
        $content = "<P>Empty page.</P>\n\n\n";

        sendto_file( "$srcDir/default.tmpl", '', $content );
    }

    # Todo - break this out into a global template somewhere
    my $editor = qq|
<FORM ACTION="$my_url" METHOD="post">
<CENTER>
<TEXTAREA NAME="d_Content" WRAP="virtual" COLS=80 ROWS=24 style="width:100%">$content</TEXTAREA>
</CENTER>
<INPUT TYPE="hidden" NAME="command" VALUE="storefile">
<INPUT TYPE="hidden" NAME="modulename" VALUE="$modulename">
<INPUT TYPE="hidden" NAME="d_Filename" VALUE="$filename">
<CENTER><INPUT TYPE="submit" VALUE=" Save Changes "></CENTER>
</FORM>
    |;
    sendto_screen( '', "Editing $filename", $editor );
}

=head2 do_single

Displays or displays an edit form (HTML) for a single record. Special handling
for admin module. Some TT2-ish stuff.

=cut

sub do_single($$$$$$$;$) {
    my ( $modulename, $command, $table, $dbh, $on, $defaults, $noticeMsg ) = @_;
    my $blank = ( $command eq 'edit' ? '' : '&nbsp;' );

    # A specific record was requested, so load that one up
    if ( $modulename eq 'users' && $defaults->{ $idField } ne $on ) {
        if ( $defaults->{ Permissions }->{ admin } ne 'Y' && $limit_view_user )
        {
            die "Cannot access record for $on\n";
        }
    }

    my $content;

    if ( $query eq 'admin'
         && -e "$srcDir/$modulename/$command.single.admin.tmpl" )
    {
        $content = loadFile( "$srcDir/$modulename/$command.single.admin.tmpl" );
    } elsif ( -e "$srcDir/$modulename/$command.single.tmpl" ) {
        $content = loadFile( "$srcDir/$modulename/$command.single.tmpl" );
    }

    my $id = $on;    # need to incorporate numbering a little more appropriately

    if ( !isInDatabase( $dbh, $table, "$idField=$id" ) ) {
        die "Couldn't find $id\n";
    }
    my $sth = selectRecords( $dbh, $table, "$idField=$id", '' );
    if ( $sth == 0 ) {

        # No available records were found
        # TODO:  Tell user we couldn't process it
        print "Couldn't process $id!\n";
    }
    my $data = $sth->fetchrow_hashref;

    if ( $content ) {
        $content = substituteValues( $content, $data,     $blank );
        $content = substituteValues( $content, $defaults, $blank );
        $content = parseTemplate(
                                  $dbh,   $content, $modulename,
                                  $table, $data
        );
        my $startTable = loadFile( "$srcDir/framed.table_tag.start.tmpl" );
        my $endTable .= loadFile( "$srcDir/framed.table_tag.end.tmpl" );

        while ( $content =~
/\[\%\s*subquery\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*([\w=]+)\s*([\!\w]+)\s*(\w+)\s*\%\]/
          )
        {
            my $htmlTable =
              makeHtmlTable( $dbh, $1, $2, $3, $4, "$5=" . $data->{ $idField },
                             $6, $7, $defaults );
            if ( $htmlTable ne '' ) {
                $htmlTable = "$startTable\n$htmlTable\n$endTable\n";
            }
            $content =~
s/\[\%\s*subquery\s*(\w+)\:\:(\w+)\s*(\w+)\s*(\w+)\s*([\w=]+)\s*([\!\w]+)\s*(\w+)\s*\%\]/$htmlTable/;
        }

    } elsif ( $command eq 'display' ) {
        my @fields = @{ getTableInfo( $dbh, $table, 'INPUT_FIELDS' ) };
        $content .= "<TABLE BORDER=0 CELLSPAING=0 CELLPADDING=2>\n";
        $content .= " <TR>\n  <TD WIDTH='150'><B><U>Field</U></B></TD>\n";
        $content .= "  <TD WIDTH='200'><B><U>Value</U></B></TD>\n </TR>\n";
        $content .=
          " <TR>\n  <TD><B>ID </B></TD>\n  <TD>$data->{uid}</TD>\n </TR>\n";
        foreach my $key ( sort @fields ) {
            my $value = '&nbsp;';
            if ( exists $data->{ $key } ) {
                $value = $data->{ $key };
            }
            $key = ucfirst( $key );
            $key =~ s/\_(\w)/ \U$1/g;
            $content .=
              " <TR>\n  <TD><B>$key</B></TD>\n  <TD>$value</TD>\n </TR>\n";
        }
        $content .= "</TABLE>\n";
    } elsif ( $command eq 'edit' ) {
        my $sname = "$srcDir/$modulename/edit.$table.form.tmpl";

        if ( $modulename eq 'projects'
             && $defaults->{ user_id } != $data->{ Lead_uid_user }
             && $defaults->{ Permissions }->{ projects_admin } ne 'Y' )
        {
            die
"You cannot edit this project's definition.  Please contact its owner or one of the administrators if the project needs updating.\n";
        }

        if ( -e $sname && $modulename ne 'projects' && $modulename ne 'users' )
        {
            $content = loadFile( $sname );

            # Substitute any dropdown boxes with dynamic data
            #           $content =~ s/\[\%\s*Hidden_info\s*\%\]/$hiddenInfo/;
            #           $content =~ s/\[\%\s*Notice\s*\%\]/$noticeMsg/;
            $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
            $content =~
              s/\[\%\s*Reqd\s*\%\]/\<FONT COLOR\=\'red\'\>\*\<\/FONT\> /g;
            $content = substituteValues( $content, \%storageError, $blank );
            $content = substituteValues( $content, $data,     $blank );
            $content = substituteValues( $content, $defaults, $blank );

            $content = processPage( $content, $data );
            $content =~ s/\[\%.*\%\]//g;

        } else {
            my %form = (
                         fields => getTableInfo(
                                                 $dbh, $table,
                                                 'INPUT_FIELDS'
                         ),
                         descriptions => getTableInfo(
                                                       $dbh, $table,
                                                       'DESCS'
                         ),
                         types  => getTableInfo( $dbh, $table, 'TYPES' ),
                         values => $data,
                         notes  => getTableInfo( $dbh, $table, 'EXTRA' )
            );
            $content = genericHtmlForm( $modulename, 'store', %form );
        }
    }
    $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
    $content =~ s/\[\%\s*Notice\s*\%\]/$noticeMsg/g;
    if ( DEBUGGING ) {
        $content =~ s/\[\%\s*(\w+)\s*\%\]/\[\% ERROR!! $1 \%\]/g;
    } else {
        $content =~ s/\[\%\s*(\w+)\s*\%\]//g;
    }
    sendto_screen( '', ucfirst( $data->{ descriptor } ), $content );
}

=head2 do_many

Displays or displays edit form or search results for many records. Special 
handling for admin. Some pseudo-TT2 stuff. (HTML)

=cut

sub do_many($$$$$$$;$) {
    my ( $modulename, $command, $table, $dbh, $on, $defaults, $noticeMsg ) = @_;
    my $blank = '&nbsp;';
    if ( $command eq 'edit' ) { $blank = ''; }

    if ( $modulename eq 'users'
         && $defaults->{ Permissions }->{ admin } ne 'Y'
         && $limit_view_user )
    {
        die "Cannot access user records\n";
    }

    # A range of items to be edited was requested

    my $rightify = ( $on =~ /\!/ );
    my $colorify = 1;
    my $newFilter = parseOn( $on );

    my $count = 0;    ###
    if ( $query eq '' ) { $query = 'all'; }

    if ( !$sort ) { $sort = $idField; }

    my @fields = @{ getTableInfo( $dbh, $table, 'FIELDS' ) };
    my $sth = getfrom_db(
                          $dbh,  $table, $on,         $filter,
                          $sort, $start, $maxitems * 2
    );
    my $content;

    my $remainder = $sth->rows() - $maxitems;
    if ( $remainder < 0 ) { $remainder = 0; }

    if ( $sth == 0 ) {
        $content =
"No available records were found, so we couldn't process the request.";
    } else {
        if ( $command eq 'edit' ) {
            $content =
              loadFile( "$srcDir/$modulename/$command.list.form.tmpl" );
        } elsif ( $command eq 'search' ) {
            $content = loadFile( "$srcDir/$modulename/display.all.page.tmpl" );

            my $header =
              loadFile( "$srcDir/$modulename/display.all.header.tmpl" );
            $content =~ s/\[\%\s*header\s*\%\]/$header/;

        } else {
            $content =
              loadFile( "$srcDir/$modulename/$command.$query.page.tmpl" );

            my $header =
              loadFile( "$srcDir/$modulename/$command.$query.header.tmpl" );
            if (
                -e "$srcDir/$modulename/$command.$query.header.$language.tmpl" )
            {
                $header =
                  loadFile(
                     "$srcDir/$modulename/$command.$query.header.$language.tmpl"
                );
            }
            $content =~ s/\[\%\s*header\s*\%\]/$header/;

        }
        my $rowTmpl = '';

        # TODO:  Make this search for default templates more thoroughly
        if ( -e "$srcDir/$modulename/$command.$query.row.tmpl" ) {
            $rowTmpl =
              loadFile( "$srcDir/$modulename/$command.$query.row.tmpl" );
        } elsif ( -e "$srcDir/$modulename/$command.row.tmpl" ) {
            $rowTmpl = loadFile( "$srcDir/$modulename/$command.row.tmpl" );
        } else {
            $rowTmpl = loadFile( "$srcDir/$modulename/display.all.row.tmpl" );
        }
        my $color   = '';
        my $rowData = '';
        my $header  = '';
        if ( -e "$srcDir/$modulename/$command.$query.header.tmpl" ) {
            $header =
              loadFile( "$srcDir/$modulename/$command.$query.header.tmpl" );
        } elsif ( -e "$srcDir/$modulename/$command.header.tmpl" ) {
            $header = loadFile( "$srcDir/$modulename/$command.header.tmpl" );
        } else {
            $header = loadFile( "$srcDir/$modulename/display.all.header.tmpl" );
        }

        my $count = 0;

        if ( $rightify && $sth->rows > $maxitems ) {
            $count = $maxitems - $sth->rows;
        }
        while ( my $data = getnext_db( $sth ) ) {
            $count++;

            # If hiliting is on, determine what color scheme to use
            if ( $hilite && defined( $data->{ $hilite } ) ) {
                if ( defined $colorScheme{ $data->{ $hilite } } ) {
                    $color = $colorScheme{ $data->{ $hilite } };
                } else {

                    # Define new color scheme
                    $color = nextColor();
                    $colorScheme{ $data->{ $hilite } } = $color;
                }
            } elsif ( $color eq "#FFFFFF" && $colorify ) {
                $color = "#CCCCCC";
            } else {
                $color = "#FFFFFF";
            }
            $data->{ row_color }  = $color;
            $data->{ modulename } = $modulename;

            last if ( $count > $maxitems );

            # Load the template for describing a row in a table, & fill it in
            if ( $count > 0 ) {
                my $rowTmpl2;
                foreach ( split /\n/, $rowTmpl ) {
                    while ( /\[\%\s*command\s+(\w+)\s*\%\]/ ) {
                        my $cmd = cgiCmd(
                                          $modulename, $1,
                                          ( on => $data->{ $idField } )
                        );
                        my $c      = $1;
                        my $action = qq|<A HREF="$cmd">|;
                        if ( -e "$htmlDir/images/$c.png" ) {
                            $action .=
                              qq|<IMG SRC="/images/$c.png" BORDER="0"></A>|;
                        } else {
                            $action .= qq|$c</A>|;
                        }
                        s/\[\% command\s+(\w+)\s*\%\]/$action/;
                    }
                    while ( /\[\%\s*(\w+)\.(\w+)\s*\%\]/ ) {
                        my $text = getFieldText( $dbh, $data, $1, $2 );

                        if ( !$text || $text eq '' ) {
                            $text = $blank;
                        }
                        s/\[\%\s*$1\.$2\s*\%\]/$text/ig;
                    }
                    $rowTmpl2 .= $_ . "\n";
                }

                # If there are any sub-tables, load those up too
                if ( $rowTmpl2 =~ /\[%\s*\w+\.[\w\.]+/ ) {
                    $rowTmpl2 =~ s/Host_type_uid\.//g;
                }

                $rowData .= substituteValues( $rowTmpl2, $data, $blank ) . "\n";
            }
        }
        $content =~ s/\[\%\s*row\s*\%\]/$rowData/;

        my $sep;
        my $pageCtrl = qq|<P ALIGN="right">|;
        if ( $start > 0 ) {
            $pageCtrl .= '<A HREF="';
            if ( $start <= $maxitems ) {
                $pageCtrl .= cgiCmd( $modulename, $command, ( start => 0 ) );
                $pageCtrl .= "\">First Page</A> ";
            } else {
                $pageCtrl .= cgiCmd(
                                   $modulename, $command,
                                   ( start => $start - $maxitems )
                );
                $pageCtrl .= "\">Prev $maxitems</A> ";
            }
            $sep = " | ";
        } else {
            $sep = " ";
        }
        if ( $remainder > 0 && ( $count >= $maxitems || $rightify ) ) {
            $pageCtrl .= $sep;
            $pageCtrl .= '<A HREF="';
            $pageCtrl .= cgiCmd(
                                 $modulename, $command,
                                 ( start => $start + $maxitems )
            );

            if ( $remainder < $maxitems ) {
                $pageCtrl .= "\">Next $remainder...</A>";
            } else {
                $pageCtrl .= "\">Next $maxitems...</A>";
            }
        }
        $pageCtrl .= "</P>\n";
        $content = $pageCtrl . $content . $pageCtrl;

        my %var = %$defaults;
        $var{ 'pageCtrl' }                 = $pageCtrl;
        $var{ 'start_table style=framed' } =
          loadFile( "$srcDir/framed.table_tag.start.tmpl" );
        $var{ end_table } = loadFile( "$srcDir/framed.table_tag.end.tmpl" );
        if ( -e "$srcDir/$modulename/$command.$query.header.tmpl" ) {
            $var{ header } =
              loadFile( "$srcDir/$modulename/$command.$query.header.tmpl" );
        } elsif ( -e "$srcDir/$modulename/$command.header.tmpl" ) {
            $var{ header } =
              loadFile( "$srcDir/$modulename/$command.header.tmpl" );
        } else {
            $var{ header } =
              loadFile( "$srcDir/$modulename/display.all.header.tmpl" );
        }
        $var{ on }         = $on;
        $var{ my_url }     = $my_url;
        $var{ modulename } = $modulename;

        $content = substituteValues( $content, \%var, $blank );
        $content = substituteValues( $content, $defaults, $blank );
        if ( DEBUGGING ) {
            $content =~ s/\[\%\s*(\w+)\s*\%\]/\[\% ERROR!! $1 \%\]/g;
        } else {
            $content =~ s/\[\%\s*(\w+)\s*\%\]//g;
        }
    }
    my $title = ucfirst( $query );
    $title =~ s/\_(\w)/ \U$1/g;

    sendto_screen( '', $title, $content );
}

=head2 do_remove

Removes items from table (by setting rsf=0; doesn\'t actually delete anything.)

=cut

sub do_remove {
    my (
         $modulename, $command,  $table,   $dbh,
         $on,         $data,     $emails,  $nextCommand,
         $nextQuery,  $defaults, $noticeMsg
      )
      = @_;
    my $content;
    my $user_id = $defaults->{ user_id } || 0;

    if ( $on !~ /^\d+$/ ) {
        $noticeMsg = "You must specify a valid ID to remove.";
    } else {
        my $id = $on;
        if ( !isInDatabase( $dbh, $table, "$idField=$id" ) ) {
            die "Couldn't find $id\n";
        }

        $data->{ $idField } = $id;
        $data->{ rsf } = 0;
        updateRecord( $modulename, $user_id, %$data );

        $table =~ s/\_/ /g;
        $noticeMsg = "$table entry $id was successfully removed";
    }
    $nextCommand = 'display';
    return ( $nextCommand, $nextQuery, $noticeMsg );
}

sub parseTemplate {
    my ( $dbh, $tmpl, $modulename, $table, $data ) = @_;
    my $output = '';

    foreach ( split /\n/, $tmpl ) {
        while ( /\[\%\s*command\s+(\w+)\s*\%\]/ ) {
            my $cmd = cgiCmd(
                              $modulename, $1,
                              ( on => $data->{ $idField } )
            );
            my $c      = $1;
            my $action = qq|<A HREF="$cmd">|;
            if ( -e "$htmlDir/images/$c.png" ) {
                $action .= qq|<IMG SRC="/images/$c.png" BORDER="0"></A>|;
            } else {
                $action .= qq|$c</A>|;
            }
            s/\[\% command\s+(\w+)\s*\%\]/$action/;
        }
        while ( /\[\%\s*(\w+)\.(\w+)\s*\%\]/ ) {
            my $text = getFieldText( $dbh, $data, $1, $2 );

            s/\[\%\s*$1\.$2\s*\%\]/$text/gi;
        }
        while ( /\[\%\s*junction\s*(\w+)\s*(\w+)\s*\%\]/ ) {
            my ( $junction, $target, $source ) = ( $1, $2, $table );

            my $id = $data->{ $idField };
            my $sth = selectRecords(
                                     $dbh,
                                     "$target<-$junction",
                                     "$junction.$source\_$idField=$id",
                                     "$junction.$target\_$idField"
            );

            my $text;
            while ( my $data = $sth->fetchrow_hashref ) {
                my $cmd = '';
                if ( exists $modules{ $target . 's' } ) {
                    $cmd =
                      cgiCmd( $target . 's', 'display',
                              ( on => $data->{ $idField } ) );
                } else {
                    $cmd =
                      cgiCmd( $target, 'display',
                              ( on => $data->{ $idField } ) );
                }
                $text .= qq|<A HREF="$cmd">|;
                $text .= $data->{ descriptor } . "</A><BR>\n";
            }
            s/\[\%\s*junction\s*$junction\s*$target\s*\%\]/$text/;
        }
        $output .= $_ . "\n";
    }

    return substituteValues( $output, $data ) . "\n";
}

########################################################################
## MAIL SUBSYSTEM                                                     ##
########################################################################

=head1 Mail subsystem

=head2 sendto_email(TO, TITLE, CONTENT)

Sends mail to $to, by a pipe to the mail command. Eidetic::Mail uses the 
Mail::Sendmail module, which works better.

=cut

sub sendto_email($$$;$) {
    my ( $to, $title, $content ) = @_;

    open( MAIL, "|mail $to -s \"$title\"" )
      || die "failure Could not open mail to $to";
    print MAIL $content;
    close MAIL;

    # TODO: Error handling
}

=head2 notifyStaff(MODULE, DATA)

Sends %data to the hardcoded staff addresses via sendto_email.

=cut

# notifyStaff
sub notifyStaff($%) {
    my ( $modulename, %data ) = @_;
    my ( $address, @addresses, $key, @keys, @lines, $sep );
    my $intro   = "$modulename entry has been submitted:\n\n";
    my $subject = "$modulename submission";

    if ( $modulename eq 'award_nomination' ) {
        @addresses = @award_nomination_emails;
        @keys = sort keys %data;    # Hook for more sophisticated sorting later
    } else {
        @addresses =
          $data{ Lab_uid } eq '1' 
          ? @project_notification_emails
          : @project_notification_emails_jp;
        if ( $data{ Coordinator_email } ne '' ) {
            push @addresses, $data{ Coordinator_email };
        }

        @keys = sort keys %data;
    }

    my $message      = '';
    my $templateFile = "$srcDir/$modulename/notifyStaff.tmpl";

    # Commented out to remove TT2 dependency for STP2 (temporarily)
    #    if (-f $templateFile) {
    #	# Load template
    #	my $template = Template->new({ABSOLUTE => 1});
    #	$template->process($templateFile, \%data, \$message)
    #	    || die "Could not process template $templateFile: ", 
    #	    $template->error(), "\n";
    #
    #    } else {
    $message = $intro;
    foreach $key ( @keys ) {

        #	    my $value = join "\n", wrap('  ', '  ', $data{$key});
        my $value = $data{ $key };
        $sep = '-' x ( length $key );
        $message .= $key . "\n" . $sep . "\n" . $value . "\n\n";
    }

    #    }

    my $to = join ', ', @addresses;
    sendto_email( $to, $subject, $message );

    return "";
}

########################################################################
## UTILITIES                                                          ##
########################################################################

# What if in the template we simply placed a notice area for plugging
# in error or success messages, and etc.?
# I need to consider how to best handle errors for instance when accessed
# via means other than HTML.
# I'd like to make this able to log errors people encounter, and be able
# to shut off reporting of errors to users entirely.

=head1 Utility functions

=head2 handleError(MODULE, ERROR)

Prints error messages; due for replacement by Eidetic::Error.

=cut

# handleError
sub handleError($$) {
    my ( $modulename, $error ) = @_;
    my ( $result, @comments ) = split ( / /, $error );
    my ( $errmsg );

    if ( !defined $result
         || $result =~ /^success/i
         || $result =~ /^message/i
         || $result =~ /^note/i
         || $result eq '' )
    {

        # do nothing; there was no error
        return '';
    } elsif ( $result eq 'failure' || $result eq 'error' ) {
        $errmsg = join ' ', @comments;
    } else {

        # This may have been an internal exception
        $result = "Error";
        $errmsg = $error;
    }
    $errmsg =~ s/at [\w\.]+ line \d+//;

    # I need a way to insert $reason and @comments into the page.

    sendto_screen( '', "$result in $modulename", $errmsg );
}

=head2 listTemplates(TYPE)

Lists available templates for $type.

=cut

# Lists available templates
sub listTemplates($) {
    my ( $type ) = @_;
    opendir( DIR, "$tmplDir/$type" )
      or die "Can't opendir $tmplDir/$type: $!";
    my @fileList = grep {/\.tmpl$/} readdir( DIR );
    closedir( DIR );
    return @fileList;
}

=head2 substituteValues(CONTENT, VARIABLES, BLANK)

This is a quick and dirty substituter; template-toolkit ought to be
used instead.  This assumes a few variables or a short content.  If
the file is long and there are many variables, this routine will be
inefficient.  Maybe I should incorporate standard default var's too?

=cut

# This is a quick and dirty substituter; template-toolkit ought to be
# used instead.  This assumes a few variables or a short content.  If
# the file is long and there are many variables, this routine will be
# inefficient.  Maybe I should incorporate standard default var's too?
sub substituteValues {
    my ( $content, $variables, $blank ) = @_;
    my ( $key,  $value, $trimValue );
    my ( $year, $month, $day ) = ( '', '', '' );
    my ( $hour, $min,   $sec ) = ( '', '', '' );

    my @months = (
                   '',    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul',
                   'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
    );
    if ( !$blank ) { $blank = ''; }

    while ( ( $key, $value ) = each( %{ $variables } ) ) {
        if ( $key ) {
            $value = $blank
              unless ( $value && $value !~ /^\s*[\.\-\_\~\?]*\s*$/ );
            if ( $key =~ /_date$/ && $value ne $blank ) {
                $value =~ /(\d\d\d\d)-(\d\d)-(\d\d) (\d\d)\:(\d\d)\:(\d\d)/;
                ( $year, $month, $day, $hour, $min, $sec ) =
                  ( $1, $2, $3, $4, $5, $6 );
                $day = ( defined $day ) ? $day + 0 : 0;
                if ( $month ) {
                    $month = $months[ $month ];
                }
                $content =~
                  s/\[\%\s*$key.date\s*\%\]/$month&nbsp;$day,&nbsp;$year/gi;
                $content =~ s/\[\%\s*$key.time\s*\%\]/$hour:$min:$sec/gi;
                $content =~ s/\[\%\s*$key\s*\%\]/$value/gi;
            } elsif ( $key =~ /_time$/ && $value ne $blank ) {
                $value =~ /[\w\-]*\s*(\d\d):(\d\d):(\d\d)/;
                my ( $hour, $min, $sec ) = ( $1, $2, $3 );
                if ( !$hour ) { $hour = 0; }
                if ( !$min )  { $min  = 0; }
                if ( !$sec )  { $sec  = 0; }
                my $ampm = 'am';
                if ( $hour > 12 ) {
                    $hour -= 12;
                    $ampm = 'pm';
                }
                $content =~ s/\[\%\s*$key\s*\%\]/$hour:$min $ampm/gi;
            } else {
                $content =~ s/\[\%\s*$key\s*\%\]/$value/gi;
            }

            if ( length $value > 40 ) {
                $trimValue = sprintf "%.37s...", $value;
                $content =~ s/\[\%\s*short\s+$key\s+(\w+)\s*\%\]/$1/gi;
                $content =~ s/\[\%\s*trim\s+$key\s*\%\]/$trimValue/gi;
            } else {
                $content =~ s/\[\%\s*short\s+$key\s+(\w+)\s*\%\]/$value/gi;
                $content =~ s/\[\%\s*trim\s+$key\s*\%\]/$value/gi;
            }
        }
    }

    return $content;
}

=head2 loadFile(FNAME)

Loads $fname, returns all lines as a string. On failure dies; if $fname 
doesn\'t exist gives a wacky HTML stack trace.

=cut

sub loadFile {
    my ( $fname ) = @_;
    if ( !-e $fname ) {
        die "$fname does not appear to exist.  <BR>"
          . ( caller( 2 ) )[ 3 ] . " @ "
          . ( caller( 1 ) )[ 2 ]
          . " called \n"
          . ( caller( 1 ) )[ 3 ] . " @ "
          . ( caller( 0 ) )[ 2 ]
          . "<BR>\n";
    }
    open( GF_FILE, "< $fname" ) || die "$fname could not be opened";
    my @lines = <GF_FILE>;
    close( GF_FILE ) || die "$fname could not be closed";
    return join '', @lines;
}

=head2 sendto_file(FNAME, TITLE, CONTENT)

Saves text to a file.

=cut

sub sendto_file($$$) {
    my ( $fname, $title, $content ) = @_;

    open FILE, "> $fname" || die "Failed to open $fname";
    if ( $title ) {
        print FILE "$title";
    }
    print FILE $content;
    close FILE;
    return 1;
}

=head2 generateRandomPassword( )

Generates a random eight character password, using letters, numbers,
and a little punctuation.

=cut

sub generateRandomPassword {
    my @chars = ( 'A' .. 'Z', 'a' .. 'z', 0 .. 9, qw(! @ ^ &) );
    return join ( "", @chars[ map { rand @chars } ( 1 .. 8 ) ] );
}

########################################################################
## TEXT TABLE SUBSYSTEM                                               ##
########################################################################

=head1 Text table subsystem

A general purpose file-oriented table loader.  Useful for
loading up data from text files when it's in a simple column oriented
layout.  In some cases this will be easier for users to deal with than
a field of widgets or a bunch of SQL.  Choice is good.

I'd like this to use syntax that is simple, easy to parse, and
straightforward to non-technical users.  If at all possible I'd
like to be able to use it for a variety of other things, but if
it at least does what we need, with a minimum of fuss, that's fine.

This is kind of a typical file usage thing...  I wonder if there are
already special commands for handling these that I should use?

=head2 loadTextTable(FNAME)

Loads'n'parses the text table in $fname; returns a hash with the data.

=cut

sub loadTextTable($) {
    my ( $fname ) = @_;
    if ( !-e $fname ) {

        # TODO - Add error handling
        return undef;
    }

    my ( @fields, $mode, $h, $i, %record, $line, $title );  # could use ordering
    $i    = 0;
    $h    = 0;
    $mode = '';
    open FILE, "< $fname";
    while ( defined( $line = <FILE> ) ) {
        chomp $line;    # Delete the \n

        # If there is a continuation character, keep appending lines
        if ( $line =~ s/\\\s*$// ) {
            $line .= <FILE>;
            redo unless eof( FILE );
        }

        $line =~ s/#.*//;     # no comments
        $line =~ s/^\s+//;    # no leading whitespace
        $line =~ s/\s+$//;    # no trailing whitespace

        # This section handles XML-ish tags
        if ( $mode eq '' && $line =~ m|^<[^/]| ) {  # line is starting a new tag
                # This is an XML-ish tag to change into a new mode
            my @a = split />/, $line;
            $mode = shift @a;
            $mode =~ s/<//;
            $line = join '>', @a;
        }
        if ( $line =~ m|</$mode>| ) {
            $line =~ s|</$mode>||;    # remove the tag, leave rest
            $mode = '';               # unset the mode
        }

        next unless length $line;    # anything left?

        if ( $mode ne '' ) {

            # We are in a mode, so slurp into the hash as appropriate
            $record{ 'text' }->{ $mode } .= $line . "\n";

        } else {

            # This section takes care of data tables
            if ( $line =~ /^\=+/ ) {

                # It's a title line; extract it
                $line =~ s/\=//g;    # delete all equals signs
                $title = $line;

                $record{ 'title' }->{ $h } = $title;

            } elsif ( $line =~ /^\[/ ) {

                # It's a header; we will redefine the field names

                # First do some regexp cleanup
                # This could probably be done more efficiently, but
                # it's clearer what's going on this way..
                $line =~ s/^\[//;            # chop off the leading '['
                $line =~ s/\]$//;            # chop off the trailing ']', if any
                $line =~ s/\s*\|\s*/\|/g;    # no space around the divider
                $line =~ s/^\s+//;           # no leading whitespace
                $line =~ s/\s+$//;           # no trailing whitespace
                @fields = split /\|/, $line;

                # Store header in data structure, so we know the ordering scheme
                $h++;
                $record{ 'header' }->{ $h } = $line;

            } else {

                # It's a regular line, so just process it
                $line =~ s/\s+/ /g;    # Simplify multispacing to one space
                my @data = split / /, $line;

                # We simply use the table title plus a row number as the key
                # Users of this subroutine can impose whatever other id system
                # they wish on the data, later.  Probably later we'd like to
                # implement a key object, with title and number data members.
                $i++;
                my $key = sprintf( "%s_%8d", $title, $i );

                # It would be nice to have a simple field validation
                # system, to warn user if the table is irregular

                # if @fields is defined, use it to define the columns
                if ( @fields ) {
                    my ( $field, $c );
                    $c = 0;

                    # Store which header this record belongs to
                    # Later on, this should be used to tell what table this
                    # belongs to.  Later still, we'll do the right thing and
                    # make sure each table is stored in a separate data
                    # structure.
                    $record{ $key }->{ 'schema' } = $h;

                    foreach $field ( @fields ) {

                        # Put the data into the recordset
                        $record{ $key }->{ $field } = ( shift @data );
                    }

                    # Add whatever data remains to the last field as a string
                    if ( @data ) {
                        $record{ $key }->{ $fields[ $#fields ] } .=
                          ' ' . ( join ' ', @data );
                    }

                } else {

                    # autoname the columns [Field1], etc.
                }
            }
        }
    }
    close FILE;

    return %record;
}

########################################################################
## CONFIGURATION SUBSYSTEM                                            ##
########################################################################

=head2 stringify(VALUE, LEVEL)

Stringifies $value, handling refs correctly; array & hash refs are indented by 
$level. Returns undef if it can\'t be stringified.

=cut

sub stringify {
    my ( $value, $level ) = @_;
    my $type = ref( $value );
    if ( !$type ) {

        # Just a scalar
        return $value;
    } elsif ( $type eq 'SCALAR' ) {

        # Scalar reference
        return stringify( $$value );
    } elsif ( $type eq 'ARRAY' ) {

        # Array reference
        my ( $item, @items );
        my $spaces = ' ' x $level;
        foreach $item ( @$value ) {
            push @items, $spaces . stringify( $item, $level + 1 );
        }
        if ( $level == 0 ) {
            return join ( ",\n", @items ) . "\n";
        } else {
            return "(\n$spaces"
              . join ( ",\n$spaces", @items )
              . "\n$spaces)\n";
        }
    } elsif ( $type eq 'HASH' ) {

        # Hash reference
        my ( $item, @items );
        my $spaces = ' ' x $level;
        while ( my ( $key, $value ) = each %$value ) {
            my $valStr = stringify( $value, $level + 1 );
            if ( $valStr ) {
                push @items, "$spaces$key = $valStr\n";
            }
        }
        if ( $level == 0 ) {
            return join ( ",\n", @items ) . "\n";
        } else {
            return "{\n$spaces"
              . join ( ",\n$spaces", @items )
              . "\n$spaces}\n";
        }
    } elsif ( $type eq 'CODE' )   {   # These we'll not do anything with for now
    } elsif ( $type eq 'REF' )    {
    } elsif ( $type eq 'GLOB' )   {
    } elsif ( $type eq 'IO' )     {
    } elsif ( $type eq 'LVALUE' ) {
    } elsif ( $type eq 'Regexp' ) {
    }
    return undef;
}

=head2 trim(OUT)

Removes leading and trailing ws from @out.

=cut

# Removes leading and trailing whitespace
sub trim(@) {
    my @out = @_;

    for ( @out ) {
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @out : $out[ 0 ];
}

########################################################################
## ADMINISTRATIVE SUBSYSTEM                                           ##
########################################################################

=head1 Administrative subsystem

=head2 do_admin(CGI, MODULE, COMMAND, ON, USER, DATA)

Displays admin form (if user has access) and handle input from it.

=cut

sub do_admin($$$$$) {
    my ( $cgi, $modulename, $command, $on, $user, $data ) = @_;

    my $title = "Administration Page";

    my $content = "Select a command ";
    my $color   = '#CCCCCC';

    $content .= loadFile( "$srcDir/framed.table_tag.start.tmpl" );

    foreach $modulename ( sort keys %modules ) {
        next if ( !$modulename || $modulename eq 'admin' );

        my $name = ucfirst( $modulename ) || '';
        $name =~ s/\_(\w)/ \U$1/g;
        if ( $color eq "#FFFFFF" ) {
            $color = "#CCCCCC";
        } else {
            $color = "#FFFFFF";
        }
        $content .= qq| <TR BGCOLOR="$color">\n  <TD><B>$name</B></TD>\n|;
        $content .= qq|  <TD><A HREF="|
          . cgiCmd( $modulename, 'display&on=0-' )
          . qq|">List</A></TD>\n|;
        $content .= qq|  <TD><A HREF="|
          . cgiCmd( $modulename, 'edit', on => '' )
          . qq|">New</A></TD>\n|;
        $content .= qq|  <TD><A HREF="|
          . cgiCmd( $modulename, 'edit&on=0-' )
          . qq|">Edit</A></TD>\n|;
    }

    # Load up list of boilerplate emails we can send out for various situations,
    # that the administrator can edit or send along.
    my @emailTemplates = listTemplates( 'email' );
    my @pageTemplates  = listTemplates( 'page' );
    my @tableTemplates = listTemplates( 'table' );

    $content .= loadFile( "$srcDir/framed.table_tag.end.tmpl" );

    sendto_screen( '', $title, $content );
}

########################################################################
## STP SUBSYSTEM                                                      ##
########################################################################

=head2 do_stp

Handles commands for the stp module.

=cut

sub do_stp {
    my ( $cgi, $modulename, $command, $on, $user, $data ) = @_;
    my ( $content, %form );
    my $noticeMsg = '';

    my $user_id = $user->{ $idField };

    my $dbh   = $db_handles{ 'eidetic' };
    my $table = 'test_request';

    $user->{ status }   = "Queued";
    $user->{ Username } = $auth_user;
    $user->{ Email }    = $user->{ Real_email };
    $user->{ user_id }  = $user->{ $idField };

    my @emails = ();

    my $commands = $modules{ $modulename }->{ commands };

    if ( $command eq 'new_test' ) {
        $on = '';
    }

    my ( $newCommand, $newQuery );
    ( $newCommand, $newQuery, $noticeMsg ) = $commands->{ $command }->(
      $modulename, $command, $table, $dbh,  $on, $data,
      \@emails,    $command, $query, $user, ''
    );
    $on = '';
    if ( $query eq 'all' && $newQuery eq '' ) {

        # skip
    } elsif ( $newCommand ne $command || $newQuery ne $query ) {
        $commands->{ $newCommand }->(
          $modulename, $newCommand, $table,   $dbh,
          $on,         $data,       \@emails, $newCommand,
          '',          $user,       $noticeMsg
        );
    }

}

########################################################################
## USERS SUBSYSTEM                                                    ##
########################################################################

=head1 Users subsystem

=head2 do_reset_password

Checks permissions, resets password and e-mails the user.

=cut

sub do_reset_password {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $user,       $noticeMsg
      )
      = @_;

    my $id = $on;

    # Consider wanting to reset a range of passwords...

    if ( $user->{ Permissions }->{ admin } ne 'Y'
         && $id != $user->{ $idField } )
    {
        die "You cannot reset other people's passwords.";
    }

    # Generate a new password
    my $passwd = generateRandomPassword();

    # Prepare email msg for user
    $dataHandle->{ Userpassword } = $passwd;
    my $message = loadFile( "$srcDir/users/password.reset.email.tmpl" );
    $message = substituteValues( $message, $dataHandle );

    my $sth = selectRecords( $dbh, $table, "$idField=$id", '' );
    my $data = $sth->fetchrow_hashref;
    if ( $data && $data->{ Real_email } && $data->{ Real_email } ne '' ) {

        # Encrypt password
        my $salt = sprintf( "\$1\$%.8s\$", generateRandomPassword() );
        $dataHandle->{ Userpassword } = md5_hex( $passwd );
        $dataHandle->{ Unixpassword } = crypt( $passwd, $salt );

        # Validate the inputs and store the data
        $dataHandle->{ $idField } = $id;
        updateRecord( $modulename, $id, %$dataHandle );
        $message = substituteValues( $message, $data );
        $message =~ s/\[\%\s*(\w+)\s*\%\]//g;
        sendto_email( $data->{ Real_email }, "OSDL User Account", $message );

        my $content =
          loadFile( "$srcDir/users/display.reset_performed.page.tmpl" );
        $content =~ s/\[\%\s*.*\s*\%\]//g;
        $content = substituteValues( $content, $data );
        sendto_screen( '', ucfirst( "Password has been reset" ), $content );
    } else {
        my $content =
          loadFile( "$srcDir/users/display.reset_failed.page.tmpl" );
        $content =~ s/\[\%\s*.*\s*\%\]//g;
        $content = substituteValues( $content, $data );
        sendto_screen( '', ucfirst( "Could not find user's email address" ),
                       $content );
    }
}

=head2 do_edit_password

Displays a password-editing HTML form.

=cut

sub do_edit_password {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $user,       $noticeMsg
      )
      = @_;

    my $id = $on;    # need to incorporate numbering a little more appropriately

    my $sname   = "$srcDir/$modulename/edit.$table.password.form.tmpl";
    my $content = loadFile( $sname );

    $content =~ s/\[\%\s*$idField\s*\%\]/$id/g;
    $content =~ s/\[\%\s*on\s*\%\]/$id/g;

    #    $content =~ s/\[\%\s*Hidden_info\s*\%\]/$hiddenInfo/;
    $content =~ s/\[\%\s*Notice\s*\%\]/$noticeMsg/;
    $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
    $content =~ s/\[\%\s*Reqd\s*\%\]/\<FONT COLOR\=\'red\'\>\*\<\/FONT\> /g;

#    $content =~ s/\[\%\s*New_Password_Error\s*\%\]/$storageError{New_Password_Error}/g;
    $content = substituteValues( $content, \%storageError );

    $content =~ s/\[\%.*\%\]//g;

    sendto_screen( '', ucfirst( "Change password" ), $content );
}

=head2 do_store_password

Checks permissions and stores the password from the edit form.

=cut

sub do_store_password {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $user,       $noticeMsg
      )
      = @_;

    $nextCommand = 'display';

    authorizeUser( $dbh, $on, $auth_user )
      || die "You do not have access to change this password.";

    my $user_id = $user->{ user_id };

    my $oldpasswd = $dataHandle->{ Old_password };
    my $passwd    = $dataHandle->{ New_password };
    my $passwd2   = $dataHandle->{ New_password_2 };

    if ( !$passwd || $passwd eq '' ) {
        $storageError{ New_Password_Error } = 'Please enter password';
        $command = 'reedit_password';
    }
    if ( !$passwd2 || $passwd2 eq '' ) {
        $storageError{ New_Password_Error } = 'Please enter password twice';
        $command = 'reedit_password';
    }
    if ( $passwd ne $passwd2 ) {

        # Error, passwords don't match
        $storageError{ New_Password_Error } = "Passwords don't match...";
        $command = 'reedit_password';
    }
    if ( md5_hex( $oldpasswd ) ne $user->{ Userpassword } ) {

        # Error, wrong password provided
        $storageError{ Old_Password_Error } = "Incorrect password...";
        $command = 'reedit_password';
    }

    if ( $command eq 'reedit_password' ) {
        return ( $command, $nextQuery, $noticeMsg );
    } else {
        my $salt = sprintf( "\$1\$%.8s\$", generateRandomPassword() );

        # Encrypt password
        $user->{ Userpassword } = md5_hex( $passwd );
        $user->{ Unixpassword } = crypt( $passwd, $salt );

        # Validate the inputs and store the data
        my %data = updateRecord( $modulename, $user_id, %$user );

        return ( $nextCommand, $nextQuery, "Updated password" );
    }
}

=head2 do_users_thanks

Displays a message thanking user for registering

=cut

sub do_users_thanks {
    my (
         $modulename, $command,    $table,   $dbh,
         $on,         $dataHandle, $emails,  $nextCommand,
         $nextQuery,  $defaults,   $noticeMsg
      )
      = @_;

    my $content;
    my $filename = "$srcDir/$modulename/display.thanks.page.tmpl";
    if ( -e $filename ) {
        $content = loadFile( $filename );
        $content =~ s/\[\%\s*my_url\s*\%\]/$my_url/g;
        $content =~ s/\[\%\s*modulename\s*\%\]/$modulename/g;
        $content =~ s/\[\%\s*.*\s*\%\]//g;

    } else {
        $content = "Thanks!  Your application has been received.";
    }

    sendto_screen( '', ucfirst( "Thanks!  You are now signed up as a user" ),
                   $content );
    return ( $nextCommand, $nextQuery, $noticeMsg );
}

=head1 Misc. stuff

=head2 rdfEntry(DATA, ITEM)

Displays a news entry in RDF format.

=cut

# Displays a news entry in RDF format
sub rdfEntry($$) {
    my ( $data, $item ) = @_;

    my $cmd = cgiCmd( $modulename, 'showEntry', ( id => $data->{ KEY } ) );
    $cmd =~ s/\&/\&amp\;/g;

    print "  <item>\n    <title>" . $data->{ $item };
    print "</title>\n    <link>";
    print $cmd . "</link>\n";

    print "  </item>\n\n";
}

=head2 checkCriteria(A, OP, B)

Performs an indicated comparison operator on values A and B, and
returns true or false as appropriate.  Operators are:

EQ, GT, LT, GE, LE, NE --  ==, >, <, >=, <=, !=  respectively

CN -- Contains

NC -- Does not contain

ST -- Starts with

EN -- Ends with

=cut

sub checkCriteria($$$) {
    my ( $a, $op, $b ) = @_;

    return 0 if ( !$op || !$a || !$b );

    return ( ( $op eq 'EQ' && $a eq $b )
             || ( $op eq 'GT' && $a gt $b )
             || ( $op eq 'LT' && $a lt $b )
             || ( $op eq 'GE' && $a ge $b )
             || ( $op eq 'LE' && $a le $b )
             || ( $op eq 'NE' && $a ne $b )
             || ( $op eq 'CN' && $a =~ /$b/ )
             || ( $op eq 'NC' && $a !~ /$b/ )
             || ( $op eq 'ST' && $a =~ /^$b/ )
             || ( $op eq 'EN' && $a =~ /$b$/ ) );
}

=head2 getQuery(MODULE, QUERY)

Returns the SQL conditions for a preset query, or the default query for a given
module if no preset is specificed.

=cut

sub getQuery($$) {
    my ( $modulename, $query ) = @_;

    my %default_queries = (
                            stp      => [ 'test_request', '', '' ],
                            projects => [ 'project',      '', '' ],
    );

# This is just a temporary kludge; later we'll need to have a registration sys...
    my %queries = (
        staff => [ 'user', 'Real_email LIKE "%osdl.org"', 'descriptor' ],

        all_queued_tests => [
                     'test_request',
                     'status="queued"||status="QUEUED"' . '&&auto_publish=1',
                     $idField . ' DESC'
        ],
        all_in_progress_tests => [
            'test_request',
            'status="running"'
            . '||status="Installing Operating System"'
            . '||status="Downloading Patch"'
            . '||status="Compiling Kernel"'
            . '||status="Rebooting After Kernel Install"'
            . '||status="Downloading Test"'
            . '||status="Unpacking Test"'
            . '||status="Running Test"'
            . '||status="Uploading Results"'
            . '||status="Notifying User"'
            . '&&auto_publish=1',
            $idField . ' DESC'
        ],
        all_completed_tests => [
            'test_request',
            'status="complete"'
            . '||status="Failed"'
            . '||status="Complete"'
            . '||status="CANCEL"'
            . '||status="CANCELLED"'
            . '&&auto_publish=1',
            $idField . ' DESC'
        ],
        queued_tests => [
                  'test_request',
                  'status="queued"||status="QUEUED"&&created_by=[%user_id%]',
                  $idField . ' DESC'
        ],
        in_progress_tests => [
            'test_request',
            'status="running"'
            . '||status="Installing Operating System"'
            . '||status="Downloading Patch"'
            . '||status="Compiling Kernel"'
            . '||status="Rebooting After Kernel Install"'
            . '||status="Downloading Test"'
            . '||status="Unpacking Test"'
            . '||status="Running Test"'
            . '||status="Uploading Results"'
            . '||status="Notifying User"'
            . '&&created_by=[%user_id%]',
            $idField . ' DESC'
        ],
        completed_tests => [
            'test_request',
            'status="complete"'
            . '||status="Failed"'
            . '||status="Complete"'
            . '||status="CANCEL"'
            . '||status="CANCELLED"'
            . '&&created_by=[%user_id%]',
            $idField . ' DESC'
        ],
        my_tests => [
                      'test_request', 'created_by=[%user_id%]',
                      $idField . ' DESC'
        ],

        us_lab_equipment => [
                'host_temp', 'lab_uid=1' . '&&room="Arctic"||room="Tropic"',
                'descriptor'
        ],
        jp_lab_equipment => [ 'host_temp', 'lab_uid=2', 'descriptor' ],

        host_test_requests => [
                                'test_request', '1=1',
                                'last_updated_date DESC'
        ],

        host_components => [ 'component_temp', '1=1', 'descriptor' ],

        coming_events => [
            'event',

            #                                             'event_type_uid=1',
            'Start_date'
        ],
        available_jobs => [ 'article', 'article_type_uid=1', 'created_date' ],
        rfqs => [ 'article', 'article_type_uid=2', 'created_date DESC' ],
        press_releases => [
                            'article', 'article_type_uid=3',
                            'created_date DESC'
        ],
        analysis_reports => [
                              'article', 'article_type_uid=4',
                              'created_date DESC'
        ],
        presentations => [
                           'article', 'article_type_uid=5',
                           'created_date DESC'
        ],
        tips        => [ 'article', 'article_type_uid=6', 'created_date DESC' ],
        latest_news => [
                         'article', 'article_type_uid=7',
                         'created_date DESC'
        ],
        linux_stp_news => [
                'article', "article_type_uid=7&&Topic='linux_stp'",
                'created_date DESC'
        ],
        gcc_stp_news => [
                  'article', "article_type_uid=7&&Topic='gcc_stp'",
                  'created_date DESC'
        ],
        stp_results => [
                         'article', 'article_type_uid=8',
                         'created_date DESC'
        ],
        kernel_testing_docs => [
                                 'article', 'article_type_uid=9',
                                 'created_date DESC'
        ],
        documents => [
                       'article', 'article_type_uid=10',
                       'created_date DESC'
        ],

        project_actions => [
                    'action', 'Status_uid_action_state <> 5',
                    'last_updated_date DESC'
        ],

        project_products => [ 'product', '1=1', 'created_date DESC' ],

        project_docs => [ 'document', '1=1', 'created_date DESC' ],

        user_docs => [ 'document', '1=1', 'created_date DESC' ],

        corp_labs => [ 'lab', '1=1', 'descriptor' ],

        lab_depts => [ 'dept', '1=1', 'descriptor' ],

        dept_projects     => [ 'project', '1=1', 'Proj_name' ],
        proposed_projects => [
            'project',
            '(Acceptance_state="Proposal" OR Acceptance_state="Evaluation")',
            'Proj_name'
        ],
        active_projects => [
                             'project', 'Acceptance_state="Approved"',
                             'Proj_name'
        ],
        completed_projects => [
            'project',
            'Acceptance_state="Closed" AND Development_state <> "Cancelled"',
            'Proj_name'
        ],
        proposed_projects_us => [
            'project',
'Lab_uid=1 AND (Acceptance_state="Proposal" OR Acceptance_state="Evaluation")',
            'Proj_name'
        ],
        active_projects_us => [
                    'project', 'Lab_uid=1 AND Acceptance_state="Approved"',
                    'Proj_name'
        ],
        completed_projects_us => [
            'project',
'Lab_uid=1 AND Acceptance_state="Closed" AND Development_state <> "Cancelled"',
            'Proj_name'
        ],
        proposed_projects_jp => [
            'project',
'Lab_uid=2 AND (Acceptance_state="Proposal" OR Acceptance_state="Evaluation")',
            'Proj_name'
        ],
        active_projects_jp => [
                    'project', 'Lab_uid=2 AND Acceptance_state="Approved"',
                    'Proj_name'
        ],
        completed_projects_jp => [
            'project',
'Lab_uid=2 AND Acceptance_state="Closed" AND Development_state <> "Cancelled"',
            'Proj_name'
        ]
    );

    if ( exists $queries{ $query } ) {
        return @{ $queries{ $query } };
    } elsif ( exists $default_queries{ $modulename } ) {
        return @{ $default_queries{ $modulename } };
    } else {
        return ( '', '', '' );
    }
}

=head2 makeHtmlTable(DBH, MODULE, COMMAND, QUERY, STYLE, FILTER, SORT, MAXITEMS, DEFAULTS)

Inserts the specified data into an HTML table and returns it.

=cut

sub makeHtmlTable {
    my (
         $dbh,       $modulename, $command,  $query,  $style,
         $addFilter, $addSort,    $maxitems, $defaults
      )
      = @_;

    my ( $table, $filter, $sort ) = getQuery( $modulename, $query );
    if ( $table eq '' ) {
        $table = $query;
        if ( $addFilter ) { $filter = $addFilter; }
    } else {
        $filter =~ s/\[\%\s*user_id\s*\%\]/$defaults->{user_id}/g;
        if ( $addFilter ) { $filter = join ( ' AND ', $filter, $addFilter ); }
    }
    if ( $addSort ) { $sort = $addSort; }
    my $pth     = "$srcDir/$modulename/$command.$query";
    my $rowTmpl = loadFile( "$pth.row.tmpl" );
    my $content = '';

    # Load data from database
    my $sth = selectRecords( $dbh, $table, $filter, $sort );

    # Insert into the template text
    my $count = 0;
    my $color = '';
    while ( $sth != 0
            && ( my $data = $sth->fetchrow_hashref )
            && $count < $maxitems )
    {
        if ( $color eq "#FFFFFF" ) {
            $color = "#CCCCCC";
        } else {
            $color = "#FFFFFF";
        }

        # Load the template for describing a row in a table, & fill it in
        my $line = substituteValues( $rowTmpl, $data, '&nbsp;' );
        $line =~ s/\[\%\s*row_color\s*\%\]/$color/g;
        while ( $line =~ /\[\% command (\w+) \%\]/ ) {
            my $cmd = cgiCmd(
                              $modulename, $1,
                              ( on => $data->{ $idField } )
            );
            my $action = qq|<A HREF="$cmd">$1</A>|;
            $line =~ s/\[\% command (\w+) \%\]/$action/;
        }
        while ( $line =~ /\[\%\s*(\w+)\.(\w+)\s*\%\]/ ) {
            my $text = getFieldText( $dbh, $data, $1, $2 ) || '';

            $line =~ s/\[\%\s*$1\.$2\s*\%\]/$text/ig;
        }
        if ( $content eq '' ) {
            $content = loadFile( "$pth.header.tmpl" );
        }
        $content .= $line;
        $count++;
    }

    return $content;
}

=head2 genericHtmlForm(DATA)

This routine creates an HTML form using the data hash provided.
The data hash must be of the following form:

my %data = ( fields => [ ],
             descriptions => { },
             types => { },
             values => { },
             notes => { }
);

The fields array specifies the sort order of the items in the form.

Form element names will be prepended with 'd_' so that Eidetic can
parse the returned output.

Adapted from code written originally by Adam Wendt <billyjoeray@netzero.net>.

=cut

sub genericHtmlForm($$%) {
    my ( $modulename, $command, %data ) = @_;
    my $text;

    if ( !$modulename ) { $modulename = ''; }

    my $rec_id = $data{ values }->{ $idField } || -1;
    $text = qq|<FORM ACTION="$my_url" METHOD="POST">\n|;
    $text .= qq|<INPUT TYPE="hidden" NAME="modulename" VALUE="$modulename">\n|;
    $text .= qq|<INPUT TYPE="hidden" NAME="command" VALUE="store">\n|;
    if ( $rec_id > 0 ) {
        $text .= qq|<INPUT TYPE="hidden" NAME="on" VALUE="$rec_id">\n|;
        $text .= qq|<INPUT TYPE="hidden" NAME="$idField" VALUE="$rec_id">\n|;
    }
    $text .= qq|<TABLE BORDER="0">\n|;

    foreach my $field ( @{ $data{ fields } } ) {
        my $type  = $data{ types }  ? $data{ types }->{ $field }  : 'hidden';
        my $value = $data{ values } ? $data{ values }->{ $field } : '';
        my $note  = $data{ notes }  ? $data{ notes }->{ $field }  : '';
        my $desc  =
          $data{ descriptions } ? $data{ descriptions }->{ $field } : '';
        my %foreign_key;

        if ( !$type )  { $type  = 'string'; }
        if ( !$value ) { $value = ''; }

        # Check for relation into another table
        # (A foreign key, if mysql supported it)
        if ( $type =~ /^int/i && $field =~ /^(\w+)\_uid/i ) {
            $type = 'foreign_key';
            $desc = ucfirst( $1 );
            $desc =~ s/\_/ /g;
        }

        # Any fields noted as auto_increment are definitely not enterable,
        # so there is zero reason to show them anywhere on the form; skip 'em.
        next if ( $note =~ /auto_increment/i );
        next if ( !$field );
        next if ( $field =~ /password/ );  # Skip fields that might be passwords

        $text .= qq| <TR>\n  <TD VALIGN="top">\n|;
        if ( $desc ) {
            $text .= qq|<b>$desc</b>\n|;
        } else {
            $text .= qq|&nbsp;\n|;
        }
        $text .= qq|  </TD>\n  <TD>\n|;

# SQL Type conversion (for flexibility...  this is somewhat hackish though
# and probably should be made much cleaner and clearer, with a deliberate
        # SQL type to Widget type converter.)
        if ( $type eq 'foreign_key' ) {
            my $table;
            if ( $field =~ /^(\w+)\_uid$/ ) {
                $table = lc( $1 );
            } elsif ( $field =~ /^(\w+)\_uid\_(\w+)$/ ) {
                $table = lc( $2 );
            }
            next if ( $table eq $person_table );

            # Next lookup the values for this table.
            my $dbname = "eidetic";
            my $dbh    = getDbHandle( $dbname );
            if ( checkTableExistance( $dbh, $table ) ) {
                my $sth = selectRecords( $dbh, $table, '', '' );
                while ( my $data = $sth->fetchrow_hashref ) {
                    $foreign_key{ $data->{ $idField } } = $data->{ descriptor };
                }
                $type = 'menu(KEYED)';
            } else {

                # The indicated uid hasn't got an associated table
                $type = 'string';
            }
        } elsif ( $type =~ /^varchar/i || $type =~ /^char/i ) {
            $type = 'string';
        } elsif ( $type =~ /^longtext/i ) {
            $type = 'memo';
        } elsif ( $type =~ /^tinyint/i ) {

            # This field is probably a select...
            $type = 'string';
        } elsif ( $type =~ /^int/i
                  || $type =~ /^smallint/i
                  || $type =~ /^numeric/i )
        {
            $type = 'string';
        } elsif ( $type =~ /^float/i
                  || $type =~ /^real/i
                  || $type =~ /^double/i )
        {
            $type = 'string';
        } elsif ( $type =~ /date/i || $type =~ /time/i ) {
            $type = 'string';
        } elsif ( $type =~ /^text$/i ) {
            $type = 'memo';
        } elsif ( $type =~ /text/i ) {
            $type = 'string';
        } elsif ( $type =~ /^enum/i ) {
            $type =~ s/enum/radio/i;
            $type =~ s/[\'\"]//g;
        } elsif ( $type =~ /boolean/i ) {
            $type = "checkbox($field)";
        }

        if ( $type eq 'string' ) {
            $text .=
qq|<input type="text" name="d_$field" value="$value" SIZE="30"><br>\n|;
        } elsif ( $type eq 'memo' ) {

            # TODO:  Add filters for if $value contains a textarea tag.
            $text .=
              qq|<textarea name="d_$field" cols=80 rows=10 wrap=virtual>|;
            $text .= qq|$value</textarea><br>\n|;
        } elsif ( $type =~ /checkbox\((.*)\)/ ) {
            $text .= qq|<input type="checkbox" name="d_$field"|;
            if ( $value ) { $text .= " checked"; }
            $text .= qq|> $1<br>\n|;
        } elsif ( $type eq 'submit' ) {
            $text .= qq|<input typ="$type" name="d_$field"|;
            if ( $value ne "" ) {
                $text .= qq| value="$value"|;
            }
            $text .= "><br>\n";
        } elsif ( $type eq 'reset' ) {
            $text .= qq|<input type="$type"> $note<br>\n|;
        } elsif ( $type eq 'file' ) {
            $text .=
              qq|<input type="$type" name="d_$field" value="$value"><br>\n|;
        } elsif ( $type =~ /^radio\((.*)\)$/ ) {
            foreach my $radio ( split ( /\,/, $1 ) ) {
                $text .= qq|<input type="radio" name="d_$field" value="$radio"|;
                if ( $value eq $radio ) {
                    $text .= " checked";
                }
                $text .= qq|> $radio\n|;
            }
        } elsif ( $type =~ /^menu\((.*)\)$/ ) {
            if ( %foreign_key ) {
                $text .= qq|<select name="d_$field"|;
                $text .= qq|>\n|;
                if ( $1 eq 'KEYED' ) {
                    while ( my ( $val, $val_nm ) = each %foreign_key ) {
                        $text .= qq|<option value="$val"|;
                        if ( $value eq $val ) { $text .= qq| selected|; }
                        $text .= qq|>$val_nm</option>\n|;
                    }
                } else {
                    foreach my $item ( split ( /\,/, $1 ) ) {
                        $text .= qq|<option|;
                        if ( $value eq $item ) { $text .= qq| selected|; }
                        $text .= qq|>$item</option>\n|;
                    }
                }
                $text .= qq|</select>\n$note<br>\n|;
            } else {
                $text .= qq|<I>No items to select from</I>\n|;
            }
        } elsif ( $type =~ /^image\((.*)\)$/ ) {
            $text .=
qq|<input type="$type" name="d_$field" src="$1" alt="$value"><br\n|;
        } elsif ( $type eq 'password' ) {
            $text .= qq|<input type="$type" name="d_$field"><br>\n|;
        } elsif ( $type eq 'hidden' ) {
            $text .=
              qq|<input type="$type" name="d_$field" value="$value"><br>\n|;
        }
        if ( $type ne 'hidden' ) {
            if ( $note ) { $text .= qq|$note<br>\n|; }
        } else {
            $text .= "\n";
        }
        $text .= qq|  </TD>\n </TR>\n|;
    }
    $text .= qq|</TABLE>\n|;
    $text .= qq|<INPUT TYPE="HIDDEN" NAME="command" VALUE="$command">\n|;
    $text .= qq|<P><INPUT TYPE="SUBMIT" NAME="op" VALUE="Send"></P>\n\n|;

    $text .= "</FORM>\n";

    return $text;
}

=head2 cleanFilename(FILENAME)

Returns $filename with all invalid filename characters stripped.

=cut

sub cleanFilename($) {
    my ( @filename ) = split /\//, $_[ 0 ];
    my $filename = pop @filename;

    $filename =~ s/[\-\s\+]/_/g;    # Change space, -, and + to _
    $filename =~ s/\\/\//g;         # Change \ to /
    $filename =~ s/[^\w\/]//g;      # Remove all punctuation except _ and /
    $filename =~ s/.*\///;          # Trim dir path out & remove /
    return $filename;
}

=head2 collectData

This routine gathers all of the cgi parameters that begin with the tag
'd_' (these indicate 'data' parameters that Eidetic should process.)

=cut

sub collectData($) {
    my ( $cgi ) = @_;
    my ( %data, $key );

    foreach $key ( $cgi->param() ) {

        # Ignore all keys except those starting with "d_"
        if ( $key =~ /^d\_/ ) {
            $key =~ s/^d\_(.+)/$1/;
            $data{ $key } = $cgi->param( "d_$key" );
            unless ( $cgi->param( "command" ) eq 'search' ) {
                $cgi->param(
                             -name  => "d_$key",
                             -value => ''
                );
            }
        } elsif ( $key =~ /^$idField/ ) {
            $data{ $key } = $cgi->param( "$key" );
        }
    }
    return %data;
}

=head2 cgiCmd

Creates a URL that contains the current cgi GET options for the program
ready for use inside an anchor tag or whatever.

=cut

sub cgiCmd($$%) {
    my ( $modulename, $command, %overrides ) = @_;
    my ( $cmd );

    die "bad module $modulename" unless ( exists $modules{ $modulename } );

    my %params = ();
    foreach my $param ( $cgi->param() ) {
        if ( $cgi->param( $param )
             && $cgi->param( $param ) ne ''
             && $param ne 'op' )
        {
            $params{ $param } = $cgi->param( $param );
        }
    }

    foreach my $param ( keys %overrides ) {
        $params{ $param } = $overrides{ $param };
    }

    $cmd = qq|$my_url?command=$command&modulename=$modulename|;
    foreach my $param ( keys %params ) {
        next if ( $param eq 'command' || $param eq 'modulename' );
        if ( !exists $eideticDefaults{ $param }
             || $params{ $param } ne $eideticDefaults{ $param } )
        {
            $cmd .= qq|&$param=$params{$param}|;
        }
    }

    return $cmd;
}

=head2 validateModule(?)

Returns false and displays an error message if $modulename doesn\'t exist; 
returns true if it does.

=cut

sub validateModule($) {
    if ( !exists $modules{ $modulename } ) {
        my $errmsg = qq|
            <P>The modulename you have provided, "$modulename" is not
            a valid modulename.</P><P>  If you reached this error message
            by clicking on an otherwise valid hyperlink, please email
            <A HREF="mailto:$webmaster">$webmaster</A> to let the
                maintainers know.|;
        sendto_screen( '', "Modulename error", $errmsg );
        return 0;
    } else {
        return 1;
    }
}

=head2 authenticate_user(DBH, FILTER)

Selects the user, from DBH, given by FILTER, and returns their row as a hash.

=cut

sub authenticate_user($$) {
    my ( $dbh, $filter ) = @_;

    # If logged in, get the user's info
    my $user;
    if ( $filter && isInDatabase( $dbh, $person_table, $filter ) ) {
        my $sth = selectRecords( $dbh, $person_table, $filter, '' );
        if ( $sth != 0 ) {
            $user = $sth->fetchrow_hashref;
        }

        # Get the permissions associated with this user, too
        my %permission;
        $sth =
          selectRecords( $dbh, 'permission', "user_uid=" . $user->{ $idField },
                         '' );
        while ( my $row = $sth->fetchrow_hashref ) {
            my $resource_tag = $row->{ resource_tag };
            if ( $row->{ access } eq 'Y' && $row->{ rsf } > 0 ) {
                $permission{ $resource_tag } = 'Y';
            } else {
                $permission{ $resource_tag } = 'N';
            }
        }
        $user->{ Permissions } = \%permission;
    }
    return $user;
}

############
### MAIN ###
############

sub main() {
    eval {

        # Collect data from form
        my ( %input_data ) = collectData( $cgi );

        # Validate
        # Check if the module name is valid
        validateModule( $modulename ) or exit;

        prepareDatabases( $modulename );

        my $dbh = getDbHandle( 'eidetic' );

        # If logged in, get the user's info
        my $user;
        if ( $auth_user
             && isInDatabase( $dbh, $person_table, "descriptor='$auth_user'" ) )
        {
            my $sth =
              selectRecords( $dbh, $person_table, "descriptor='$auth_user'",
                             '' );
            if ( $sth != 0 ) {
                $user = $sth->fetchrow_hashref;
            }

            # Get the permissions associated with this user, too
            my %permission;
            $sth =
              selectRecords( $dbh, 'permission',
                             "user_uid=" . $user->{ $idField }, '' );
            while ( my $row = $sth->fetchrow_hashref ) {
                my $resource_tag = $row->{ resource_tag };
                if ( $row->{ access } eq 'Y' && $row->{ rsf } > 0 ) {
                    $permission{ $resource_tag } = 'Y';
                } else {
                    $permission{ $resource_tag } = 'N';
                }
            }
            $permission{ user } = 'Y';
            $user->{ Permissions } = \%permission;
        }

        # Use dispatcher to activate relevant function
        dispatch( $cgi, $modulename, $command, $on, $user, \%input_data );

        # The variables we should be able to give to dispatch should include
        #   * The data to be presented
        #   * The desired format(s) to render into (returns a string)
        #   * The look and feel to use
        #   * The place(s) to stream to (given a string)

        closeDatabases ();

    };    # End of Eval
    handleError( $modulename, $@ );

}
1;
