# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
package STP::ASP::Host;

use STP::Validation::Suite;
use STP::DB::Host;
use STP::DB::Test;
use STP::DB::HostType;
use STP::XML::HostType;
use STP::XML::Transport;
use STP::Util::Config;
use STP::Util::Log;

require Exporter;
@ISA = qw( Exporter );

@EXPORT = qw( get_hosttypes
  get_hostinfo
  get_hosttypes_for_test
);

my $log = new STP::Util::Log( ">>/tmp/stp.log" );

# routine for returning a list of hosttypes, encapsulated in a Transport object
sub get_hosttypes {

    # shift away the ASP piece
    shift;

    # validate our arguments
    my $validation = STP::Validation::Suite::validate( "get_hosttypes", @_ );

    # is everything kosher?
    if ( $validation == 0 ) {
        $log->log( "validation failed" );
        return 0;
    }

    $log->log( "validation succeeded" );

    # get all of the host types
    my $config = STP::Util::Config::getConfig();
    my $ht     = new STP::DB::HostType( $config );

    my @hosts = $ht->getAllHostTypes();

    # convert them to xml and dump them into a transport module
    my $transport = new STP::XML::Transport();

    # iterate through the results
    foreach my $host ( @hosts ) {
        my $xml = $host->toXML( 0 );

        $transport->addDataset( "hosttype", $xml->toString() );
    }

    return $transport->toString();
}

# routine for returning a list of hosttypes, encapsulated in a Transport object
# by the test uid
sub get_hosttypes_for_test {

    # shift away the ASP piece
    shift;

    # validate our arguments
    my $validation =
      STP::Validation::Suite::validate( "get_hosttypes_for_test", @_ );

    # is everything kosher?
    if ( $validation == 0 ) {
        $log->log( "validation failed" );
        return 0;
    }

    $log->log( "validation succeeded" );
    my ( $test ) = @_;

    $log->log( "test = $test" );

    # get all of the host types
    my $config = STP::Util::Config::getConfig();
    my $ht     = new STP::DB::HostType( $config );
    $log->log( "got ht" );
    my @hosts = $ht->getAllHostTypesForTest( $test );

    $log->log( "got hosts for test" );

    # convert them to xml and dump them into a transport module
    my $transport = new STP::XML::Transport();

    # iterate through the results
    foreach my $host ( @hosts ) {
        my $xml = $host->toXML( 0 );

        $transport->addDataset( "hosttype", $xml->toString() );
    }

    $log->log( "transport = " . $transport->toString() );
    return $transport->toString();
}

sub get_hostinfo {

    # shift away the ASP piece
    shift;

    # validate our arguments
    my $validation = STP::Validation::Suite::validate( "get_hostinfo", @_ );

    # is everything kosher?
    if ( $validation == 0 ) {
        return 0;
    }

    # get all of the host types
    my $config = STP::Util::Config::getConfig();
    my $ht     = new STP::DB::HostType( $config );

    my @hosts = $ht->getAllHostTypes();

    # convert them to xml and dump them into a transport module
    my $transport = new STP::XML::Transport();

    # iterate through the results
    foreach my $host ( @hosts ) {
        my $xml = $host->toXML();

        $transport->addDataset( "hosttype", $xml->toString() );
    }

    return $transport->toString();
}

1;
