# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
package STP::ASP::Test;

use STP::Validation::Suite;
use STP::DB::Database;
use STP::DB::Test;
use STP::DB::TestRequest;
use STP::XML::Transport;
use STP::Util::Config;
use STP::Util::Log;
use POSIX qw(strftime);

require Exporter;
@ISA = qw( Exporter );

@EXPORT = qw( add_test_request
  get_tests
  get_test
  get_test_requests_for_user
);

sub add_test_request {

    # shift away the ASP piece
    shift;

    # validate our arguments
    #    my $validation = STP::Validation::Suite::validate( "add_test", @_ );

    # is everything kosher?
    #    if ( $validation == 0 )
    #    {
    #	return 0;
    #    }

    my (
         $user,          $pass,          $description,  $test_uid,
         $script_param,  $patch_tag,     $auto_publish, $lilo,
         $test_priority, $host_type_uid, $image_uid
      )
      = @_;

    # figure out the parameters to pass to writeTestRequest
    my $config = STP::Util::Config::getConfig();

    my $u = new STP::DB::User( $config );
    $u->getUserByUsername( $user );
    my $useruid = $u->getProperty( 0, "nickname" );
    my $tr = new STP::DB::TestRequest( $config );
    my $created_date = strftime( "%Y-%m-%d %H:%M:%S", localtime );
    my $uid = $tr->writeTestRequest(
                                     0,              $description,
                                     $user,          $created_date,
                                     $useruid,       $created_date,
                                     "QUEUED",       $test_uid,
                                     $script_param,  $patch_tag,
                                     $auto_publish,  $lilo,
                                     0,              $test_priority,
                                     $host_type_uid, $image_uid
    );

    my $log = new STP::Util::Log( ">>/tmp/stp.log" );
    $log->log( "host_type_uid = $host_type_uid" );

    return $uid;
}

sub get_tests {
    my $log = new STP::Util::Log( ">>/tmp/stp.log" );

    # shift away the ASP piece
    shift;

    # validate our arguments
    my $validation = STP::Validation::Suite::validate( "get_tests", @_ );

    # is everything kosher?
    if ( $validation == 0 ) {
        $log->log( "validation returned 0" );
        return 0;
    }

    my ( $test ) = @_;

    # get all of the host types
    $log->log( "loading config\n" );
    my $config = STP::Util::Config::getConfig();
    my $t      = new STP::DB::Test( $config );

    my @tests = $t->getAllTests();
    $log->log( "got all tests" );

    # convert them to xml and dump them into a transport module
    my $transport = new STP::XML::Transport();

    # iterate through the results
    foreach my $test ( @tests ) {
        my $xml = $test->toXML( 0 );

        $transport->addDataset( "test", $xml->toString() );
    }

    return $transport->toString();
}

sub get_test {
    my $log = new STP::Util::Log( ">>/tmp/stp.log" );

    # shift away the ASP piece
    shift;

    # validate our arguments
    #my $validation = STP::Validation::Suite::validate( "get_tests", @_ );

    # is everything kosher?
    #if ( $validation == 0 )
    #{
    #	$log->log( "validation returned 0" );
    #	return 0;
    #}

    my ( $test ) = @_;

    # get all of the host types

    my $config = STP::Util::Config::getConfig();
    my $t      = new STP::DB::Test( $config );

    $t->getTestByDescriptor( $test );
    if ( $t->getNumRows() == 0 ) {
        return "";
    }

    my $txml = $t->toXML( 0 );

    return $txml->toString();
}

sub get_test_requests_for_user {
    my $log = new STP::Util::Log( ">>/tmp/stp.log" );

    # shift away the ASP piece
    shift;

    # validate our arguments
#    my $validation = STP::Validation::Suite::validate( "get_test_requests_for_user", @_ );

    # is everything kosher?
    #   if ( $validation == 0 )
    #   {
    #	$log->log( "validation returned 0" );
    #	return 0;
    #    }

    my ( $uid ) = @_;

    # get all of the host types
    $log->log( "loading config\n" );
    my $config = STP::Util::Config::getConfig();
    my $tr     = new STP::DB::TestRequest( $config );

    my @tests = $tr->getTestRequestsForUid( $uid );
    $log->log( "got all tests" );

    # convert them to xml and dump them into a transport module
    my $transport = new STP::XML::Transport();

    # iterate through the results
    foreach my $test ( @tests ) {
        my $xml = $test->toXML( 0 );

        $transport->addDataset( "testrequest", $xml->toString() );
    }

    return $transport->toString();
}

1;
