# Copyright (C) 2002 Open Source Development Lab
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# This package gives access to patch_tag in database.
#

=head1 Package STP::DB::PatchTag

  uses and inherits DB::Database

=cut

package STP::DB::PatchTag;
use STP::DB::Database;

@ISA = qw( STP::DB::Database );

use strict;
use warnings;

=head1 FUNCTION new

if called with a uid, returns the patch tag for that uid
Otherwise return the empty object

=cut

sub new {
    my $self = {};
    my $package_name = shift;
    bless $self, $package_name;


    ( $self->{ config } ) = $_[0];

    $self::dbh     = "";
    $self::db_lock = 0;

    $self->{ tableName } = " patch_tag ";

    $self->{ fields } = "uid, descriptor" ;

    my $uid=$_[1];
    if ($uid){
	$self->getPatchTag($uid);
    }

    return $self;

}

=head1 FUNCTION getPatchTag

accessor function, takes uid as input

=cut

sub getPatchTag {
    my ( $self, $uid ) = @_;
    # Check that uid is digits
    if ( $uid =~ m/^\d+$/ ) {

    	my $query = qq(
                   SELECT $self->{ fields }
                   FROM $self->{ tableName }
                   WHERE uid = $uid
                   );

    	$self->{ resultSet } = $self->_select( $query );
    }
}

return 1;
