# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Result Set
#
# Author: Jerry Sievert
# Date: 10/29/01
#
# Result Set module, this is what's returned from a Database query.

package STP::DB::ResultSet;

use strict;
use warnings;
use DBI;

BEGIN { }

=head1 NAME

Set of methods for processing database queries
returns hashref containing the result set

=over

=item Functions

 new - returns hashref
 setRow - copies new row
 setProperty - could die,or return "" or return value of requested property
 getNumRows - returns { numRows }
 setNumRows - copies input to { numRows } - no return value


=back

=head2 TODO

Use the proper form of bless, sort out the error return

=cut 
    
sub new {
    my $self = {};
    bless $self;

    shift;

    my ( $stmt ) = @_;

    # see if we're instanciating without any arguments
    if ( !defined( $stmt ) ) {
        $self->{ numRows } = 0;
        return $self;
    }

    # load the data from the database handle provided
    my @rows;
    $self->{ numRows } = $stmt->rows;

    for ( my $i = 0; $i < $self->{ numRows }; $i++ ) {
        my $hash = $stmt->fetchrow_hashref;
        $self->setRow( $i, $hash );
    }

    return $self;
}

sub setRow {
    my ( $self, $row, $hash ) = @_;

    $self->{ $row } = $hash;
}

sub getProperty {
    my ( $self, $row, $property ) = @_;

    if ( $row > $self->{ numRows } || $row < 0 ) {
        die "$row is outside of range (< 0 or > " . $self->{ numRows } . ")";
    }
    if ( !defined( $self->{ $row }->{ $property } ) ) {
        return "";
    }

    return $self->{ $row }->{ $property };
}

sub getNumRows {
    my ( $self ) = @_;

    return $self->{ numRows };
}

sub setNumRows {
    my ( $self, $numRows ) = @_;

    $self->{ numRows } = $numRows;
}

END { }

1;
