# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# validation test suite
# basic tests go in here, such as an argument count test

package STP::Validation::Basic;

use STP::Validation::Suite;

my $argCount = {};

# here we have a list of the names of functions
# with the expected number of arguments for each

$argCount->{ "get_hosttypes" }          = 0;
$argCount->{ "get_hostinfo" }           = 1;
$argCount->{ "get_hosts_for_test" }     = 1;
$argCount->{ "get_hosttypes_for_test" } = 1;
$argCount->{ "get_tests" }              = 0;

# TestRequest.pm
$argCount->{ "get_test_request_for_host" } = 1;
$argCount->{ "set_host_state" }            = 2;
$argCount->{ "get_host_state" }            = 1;
$argCount->{ "set_test_request_status" }   = 2;
$argCount->{ "upload_test_results" }       = 3;

# check the number of arguments
sub argumentCountTest {
    my $name = shift;

    # check to see if it exists
    if ( !defined( $argCount->{ $name } ) ) {
        STP::Validation::Suite::setFailureMessage(
               "argumentTest called for " . "unregistered function: " . $name );
        return 0;
    }

    my $expectedCount = $argCount->{ $name };

    if ( scalar( @_ ) == $expectedCount ) {
        return 1;
    }

    # set our failure state
    STP::Validation::Suite::setFailureMessage(
        "argument count mismatch: expected "
        . $expectedCount . ", got "
        . $argCount->{ $name } );
    return 0;
}

1;
