# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# an XML object for the main settings, including methods for direct data
# interaction

package STP::XML::Config;
@ISA = qw( STP::XML::QDXml );

use strict;
use warnings;
use STP::XML::QDXml;
use STP::XML::QDXmlElement;

# config:
#
# <config>
#    <basepath></basepath>
#    <email></email>
#    <database>
#       <dsn></dsn>
#       <username></username>
#       <password></password>
#    </database>
#    <utility>
#       <power></power>
#       <image></image>
#       <ping></ping>
#    </utility>
#    <service>
#       <uri></uri>
#       <proxy></proxy>
#    </service>
#    <test>
#       <webpath></webpath>
#       <resultpath></resultpath>
#	<resulturl></resulturl>
#    </test>
#    <report>
#       <datapath></datapath>
#       <webpath></webpath>
#       <filepath></filepath>
#    </report>
# </config>

# most of this new is a copy of QDXml -- i hate perl 5's "object" model
sub new {

    # if we get a config file name passed, use it
    shift;
    my ( $configfile ) = @_;

    # instanciate the data, fill in the values
    my $self = {};
    bless $self;

    # create an array to hold the data
    $self->{ data } = [];

    # set our element name
    $self->{ elementName } = "config";

    # set up our config
    $self->addElement( "basepath", "" );
    $self->addElement( "email",    "" );

    # our database section
    my $sub1 = new STP::XML::QDXml( "database" );

    $sub1->addElement( "dsn",      "" );
    $sub1->addElement( "username", "" );
    $sub1->addElement( "password", "" );

    $self->addElement( $sub1 );

    # our utility section
    my $sub2 = new STP::XML::QDXml( "utility" );

    $sub2->addElement( "power", "" );
    $sub2->addElement( "image", "" );
    $sub2->addElement( "ping",  "" );

    $self->addElement( $sub2 );

    my $sub3 = new STP::XML::QDXml( "service" );

    $sub3->addElement( "uri",   "" );
    $sub3->addElement( "proxy", "" );

    $self->addElement( $sub3 );

    my $sub4 = new STP::XML::QDXml( "test" );

    $sub4->addElement( "webpath",    "" );
    $sub4->addElement( "resultpath", "" );
    $sub4->addElement( "resulturl", "" );
    $self->addElement( $sub4 );

    my $sub5 = new STP::XML::QDXml( "report" );

    $sub5->addElement( "datapath", "" );
    $sub5->addElement( "webpath",  "" );
    $sub5->addElement( "filepath", "" );
    $self->addElement( $sub5 );

    # if we are passed a config file to read, read it
    if ( defined( $configfile ) ) {
        $self->readFromDisk( $configfile );
    }

    return $self;
}

# export the xml object to disk, writing a file in the base directory
sub writeToDisk {
    my ( $self ) = @_;

    # get our basepath from ourselves
    my $filepath = $self->getElementValue( "basepath" );
    $filepath .= "/etc/stp/config.xml";

    # write the file
    open( FILE, "> $filepath" );
    print FILE $self->toString();
    close( FILE );
}

sub readFromDisk {
    my ( $self, $configfile ) = @_;

    # does the file exist?
    if ( !( defined( $configfile ) && ( -e $configfile ) && ( -r $configfile ) )
      )
    {
        return;
    }

    # buffer to store it all
    my $buf = "";

    open( F, $configfile );

    while ( <F> ) {
        $buf .= $_;
    }

    close( F );

    $self->parseXMLData( $buf );
}

sub get {
    my $self = shift;

    return $self->getProperty( @_ );
}

1;
