# Copyright (C) 2002 Open Source Development Lab 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# an XML object for the main settings, including methods for direct data
# interaction

package STP::XML::TestRequest;
@ISA = qw( STP::XML::QDXml );

use strict;
use warnings;
use STP::XML::QDXml;
use STP::XML::QDXmlElement;
use STP::DB::PatchTag;
use STP::DB::DistroTag;

# config:
#
# <testrequest>
#    <uid></uid>
#    <test></test>
#    <distro_tag></distro_tag>
#    <distro_name></distro_name>
#    <lilo></lilo>
#    <script_param></script_param>
#    <auto_publish><auto_/publish>
#    <patch_tag></patch_tag>
#    <patch_name></patch_name>
#    <email></email>
#    <status></status>
#    <completion_date></completion_date>
#    <host></host>
# </testrequest>

# most of this new is a copy of QDXml -- i hate perl 5's "object" model
sub new {

    # if we get a config file name passed, use it
    shift;
    my ( $configfile ) = @_;

    # instanciate the data, fill in the values
    my $self = {};
    bless $self;

    # create an array to hold the data
    $self->{ data } = [];

    # set our element name
    $self->{ elementName } = "testrequest";

    # set up our config
    $self->addElement( "uid",             "" );
    $self->addElement( "test",            "" );
    $self->addElement( "lilo",            "" );
    $self->addElement( "distro_tag",       "" );
    $self->addElement( "distro_name",       "" );
    $self->addElement( "script_param",    "" );
    $self->addElement( "auto_publish",    "" );
    $self->addElement( "patch_tag",       "" );
    $self->addElement( "patch_name",      "" );
    $self->addElement( "email",           "" );
    $self->addElement( "status",          "" );
    $self->addElement( "completion_date", "" );
    $self->addElement( "host",            "" );

    return $self;
}

sub setPatchName{
    my $this =shift;
    my $configfile = shift;

    my $patch_id = $this->getElementValue("patch_tag");

    if ($patch_id)
    {
	my $patchTag = new STP::DB::PatchTag($configfile, $patch_id);
	$this->setElementValue("patch_name", $patchTag->getProperty(0, "descriptor"));
    }

}

sub setDistroName{
    my $this =shift;
    my $configfile = shift;

    my $distro_tag = $this->getElementValue("distro_tag");

    if ($distro_tag)
    {
	my $distroName = new STP::DB::DistroTag($configfile, $distro_tag);
	$this->setElementValue("distro_name", $distroName->getProperty(0, "descriptor"));
    }

}

1;
