#!/usr/bin/perl -w 

use strict;
use AppConfig qw(:expand :argcount);

use DB::Handle;
use BASE::Log;
use STP;

my %content;
my $log = getLog( "report" );
my $cfg = getConfig();
my $db  = getDBHandle();

$db->connect;
report_daily();
report_oldest();
report_host_status();

#report_stat_per_host();
$db->disconnect;

print "\n";

sub report_daily {
    print "\n----------------------------------\n";
    print "Automated Daily Information Report\n\n";
    print `date`;
    print "----------------------------------\n\n";

    print "Last completed test: "
      . ${ $db->get( "*", "test_request",
             "Status='completed' ORDER BY Completion_date " . "DESC LIMIT 1" ) }
      { Completion_date } . "\n";

    print "Last failed test:    "
      . ${ $db->get( "*", "test_request",
                "Status='failed' ORDER BY Completion_date " . "DESC LIMIT 1" ) }
      { Completion_date } . "\n";

    print "
		Scalable Test Platform (STP) Daily Test Report

Date		Queue		Requested	Failed  	Completed
------------------------------------------------------------------------------
";

    for ( my $x = -14; $x <= 0; $x++ ) {
        my $today = `date`;
        chomp $today;
        my $now = `date -d "($today) - $x day" +%Y-%m-%d`;
        chomp $now;

        $content{ req }  = 0;
        $content{ fail } = 0;
        $content{ done } = 0;

        my @queue =
          @{ $db->getAll( "*", "test_request",
                          "created_date < '$now 00:00:00' AND "
                          . "( Completion_date > '$now 00:00:00' OR "
                          . "  Status = 'queued' OR Status = 'running' ) " ) };
        $content{ queue } = @queue;

        for ( @{ $db->getAll( "*", "test_request" ) } ) {
            if ( ${ $_ }{ created_date } =~ /$now/ ) { $content{ req }++ }

            if ( !defined( ${ $_ }{ Completion_date } )
                 || !( ${ $_ }{ Completion_date } =~ /$now/ ) )
            {
                next;
            }

            if ( ${ $_ }{ Status } =~ /fail/ ) {
                $content{ fail }++;
            }

            if ( ${ $_ }{ Status } =~ /completed/ ) {
                $content{ done }++;
            }
        }

        $content{ date } = `date -d "($today) - $x day" +%D`;
        chomp $content{ date };

        print $content{ date } . " " x  ( 16 - length( $content{ date } ) );
        print $content{ queue } . " " x ( 16 - length( $content{ queue } ) );
        print $content{ req } . " " x   ( 16 - length( $content{ req } ) );
        print $content{ fail } . " " x  ( 16 - length( $content{ fail } ) );
        print $content{ done };
        print "\n";
    }
}

sub report_oldest {
    print "\n
		Scalable Test Platform (STP) Oldest 5 Tests Report

Test		Requested		Status		Created_by
------------------------------------------------------------------------------
";

    for (
        @{ $db->getAll( "*", "test_request",
                        "Status='queued' OR Status='running' "
                        . "ORDER BY created_date LIMIT 5" ) } )
    {
        $content{ uid }    = ${ $_ }{ uid }          || "ERROR";
        $content{ req }    = ${ $_ }{ created_date } || "ERROR";
        $content{ Status } = ${ $_ }{ Status }       || "ERROR";
        $content{ user }   =
          ${ $db->get( "*", "user", "uid=" . ${ $_ }{ created_by } ) }
          { Real_email };

        print $content{ uid } . " " x    ( 16 - length( $content{ uid } ) );
        print $content{ req } . " " x    ( 24 - length( $content{ req } ) );
        print $content{ Status } . " " x ( 16 - length( $content{ Status } ) );
        print $content{ user };
        print "\n";
    }
}

sub report_host_status {
    my %host;

    for ( @{ $db->getAll( "*", "test_request", "Status='queued'" ) } ) {
        $host{ ${ $_ }{ Host_type_uid } }++;
    }

    print "\n
		Host Status Report

Host		Type	Status		Queue	Last Completed	  Last Failed
------------------------------------------------------------------------------
";

    for ( @{ $db->getAll( "*", "host", "uid < 30 ORDER BY descriptor" ) } ) {
        $content{ queue } = 0;
        if ( defined( $host{ ${ $_ }{ Host_type_uid } } ) ) {
            $content{ queue } = $host{ ${ $_ }{ Host_type_uid } };
        }

        $content{ name } =
          ${ $db->get( "*", "host", "uid=${$_}{uid}" ) }{ descriptor };

        $content{ uid } = ${ $_ }{ uid } || "ERROR";

        $content{ type } =
          ${ $db->get( "*", "host_type", "uid=" . ${ $_ }{ Host_type_uid } ) }
          { descriptor };

        $content{ status } =
          ${ $db->get( "*", "host_state", "uid=" . ${ $_ }{ Host_state_uid } ) }
          { descriptor };

        my $data =
          $db->get( "*", "test_request", "Host_uid=$content{uid} AND "
                 . "Status='completed' ORDER BY Completion_date DESC LIMIT 1" );
        if ( defined( $data ) ) {
            ${ $data }{ Completion_date } =~ /(\S+)/;
            $content{ last_date } = $1;
        } else {
            $content{ last_date } = "Never";
        }

        $data =
          $db->get( "*", "test_request", "Host_uid=$content{uid} AND "
                    . "Status='failed' ORDER BY Completion_date DESC LIMIT 1" );
        if ( defined( $data ) ) {
            ${ $data }{ Completion_date } =~ /(\S+)/;
            $content{ last_fail } = $1;
        } else {
            $content{ last_fail } = "Never";
        }

        print $content{ name } . " " x   ( 16 - length( $content{ name } ) );
        print $content{ type } . " " x   ( 8 - length( $content{ type } ) );
        print $content{ status } . " " x ( 16 - length( $content{ status } ) );
        print $content{ queue } . " " x  ( 8 - length( $content{ queue } ) );
        print $content{ last_date }
          . " " x ( 18 - length( $content{ last_date } ) );
        print $content{ last_fail };
        print "\n";
    }
}

sub report_stat_per_host {
    chomp( my $today = `date` );
    chomp( my $since = `date -d "($today) -2month" +%Y-%m-%d` );

    print "\n
		Per-Host Completed/Failed Tests Results Info

Key: Failed / Completed / Failure %

Host		T-Completed	L-Completed	T-Failed	L-Failed
------------------------------------------------------------------------------
";

    for ( @{ $db->getAll( "*", "host", "uid < 30" ) } ) {
        my %c;    # Content Holder
        my $uid = ${ $_ }{ uid };
        my $h   = ${ $db->get( "*", "host", "uid=$uid" ) }{ descriptor };

        for (
            @{ $db->getAll( "*", "test_request",
                            "(Status='failed' OR "
                            . "Status='completed') AND Host_uid=$uid AND "
                            . "Completion_date > '$since 00:00:00' "
                            . "ORDER BY Completion_date DESC" ) } )
        {
            if ( ${ $_ }{ Status } =~ /completed/ ) {
                $c{ c_l }++;
            } else {
                $c{ f_l }++;
            }
        }

        for (
            @{ $db->getAll( "*", "test_request",
                            "(Status='failed' OR "
                            . "Status='completed') AND Host_uid=$uid" ) } )
        {
            if ( ${ $_ }{ Status } =~ /completed/ ) {
                $c{ c_t }++;
            } else {
                $c{ f_t }++;
            }
        }

        unless ( defined $c{ c_t } ) { $c{ c_t } = 0 }
        unless ( defined $c{ c_l } ) { $c{ c_l } = 0 }
        unless ( defined $c{ f_t } ) { $c{ f_t } = 0 }
        unless ( defined $c{ f_l } ) { $c{ f_l } = 0 }

        print $h . " " x ( 16 - length( $h ) );
        print $c{ c_t } . " " x ( 16 - length( $c{ c_t } ) );
        print $c{ c_l } . " " x ( 16 - length( $c{ c_l } ) );
        print $c{ f_t } . " " x ( 16 - length( $c{ f_t } ) );
        print $c{ f_l } . " " x ( 16 - length( $c{ f_l } ) );
        print "\n";
    }

}

